# pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements, line-too-long

""" Generates Java classes from JSON Structure schema """
import json
import os
from typing import Dict, List, Tuple, Union, Set, Optional, Any
from avrotize.constants import JACKSON_VERSION

from avrotize.common import pascal, camel, process_template

INDENT = '    '

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None


def flatten_type_name(name: str) -> str:
    """Strips the namespace from a name"""
    if name.endswith('[]'):
        return flatten_type_name(name[:-2]+'Array')
    base_name = pascal(name.replace(' ', '').split('.')[-1].replace('>', '').replace('<', '').replace(',', ''))
    return base_name


def is_java_reserved_word(word: str) -> bool:
    """Checks if a word is a Java reserved word"""
    reserved_words = [
        'abstract', 'assert', 'boolean', 'break', 'byte', 'case', 'catch', 'char', 'class', 'const',
        'continue', 'default', 'do', 'double', 'else', 'enum', 'extends', 'final', 'finally', 'float',
        'for', 'goto', 'if', 'implements', 'import', 'instanceof', 'int', 'interface', 'long', 'native',
        'new', 'package', 'private', 'protected', 'public', 'return', 'short', 'static', 'strictfp',
        'super', 'switch', 'synchronized', 'this', 'throw', 'throws', 'transient', 'try', 'void', 'volatile',
        'while', 'true', 'false', 'null', 'record', 
    ]
    return word in reserved_words


class StructureToJava:
    """Converts JSON Structure schema to Java classes, including Jackson annotations"""

    def __init__(self, base_package: str = '') -> None:
        self.base_package = base_package.replace('.', '/')
        self.output_dir = os.getcwd()
        self.jackson_annotations = True  # Always use Jackson annotations for JSON Structure
        self.pascal_properties = False
        self.generated_types_structure_namespace: Dict[str,str] = {}
        self.generated_types_java_package: Dict[str,str] = {}
        self.schema_doc: JsonNode = None
        self.schema_registry: Dict[str, Dict] = {}

    def qualified_name(self, package: str, name: str) -> str:
        """Concatenates package and name using a dot separator"""
        slash_package_name = package.replace('.', '/')
        safe_package_slash = self.safe_package(slash_package_name.lower())
        safe_package = safe_package_slash.replace('/', '.')
        return f"{safe_package}.{name}" if package else name
    
    def join_packages(self, parent_package: str, package: str) -> str:
        """Joins package and name using a dot separator"""
        if parent_package and package:
            return f"{parent_package}.{package}".lower()
        elif parent_package:
            return parent_package.lower()
        return package.lower()

    class JavaType:
        """Java type definition"""

        def __init__(self, type_name: str, union_types: List['StructureToJava.JavaType'] | None = None, is_class: bool = False, is_enum: bool = False) -> None:
            self.type_name = type_name
            self.union_types = union_types
            self.is_class = is_class
            self.is_enum = is_enum

    def safe_identifier(self, name: str, class_name: str = '') -> str:
        """Converts a name to a safe Java identifier"""
        if is_java_reserved_word(name):
            return f"_{name}"
        if class_name and name == class_name:
            return f"{name}_"
        return name

    def safe_package(self, packageName: str) -> str:
        """Converts a name to a safe Java identifier by checking each path segment"""
        segments = packageName.split('/')
        safe_segments = [
            self.safe_identifier(segment)
            for segment in segments
        ]
        
        return '/'.join(safe_segments)
    
    def map_primitive_to_java(self, structure_type: str, is_optional: bool) -> JavaType:
        """Maps JSON Structure primitive types to Java types"""
        optional_mapping = {
            'null': 'Void',
            'boolean': 'Boolean',
            'string': 'String',
            'integer': 'Integer',
            'number': 'Double',
            'int8': 'Byte',
            'uint8': 'Short',  # Java doesn't have unsigned byte, use Short
            'int16': 'Short',
            'uint16': 'Integer',  # Java doesn't have unsigned short, use Integer
            'int32': 'Integer',
            'uint32': 'Long',  # Java doesn't have unsigned int, use Long
            'int64': 'Long',
            'uint64': 'BigInteger',  # Use BigInteger for uint64 to handle full range in JSON
            'int128': 'BigInteger',
            'uint128': 'BigInteger',
            'float8': 'Float',
            'float': 'Float',
            'double': 'Double',
            'binary32': 'Float',
            'binary64': 'Double',
            'decimal': 'BigDecimal',
            'binary': 'byte[]',
            'date': 'LocalDate',
            'time': 'LocalTime',
            'datetime': 'Instant',
            'timestamp': 'Instant',
            'duration': 'Duration',
            'uuid': 'UUID',
            'uri': 'URI',
            'jsonpointer': 'String',
            'any': 'Object'
        }
        required_mapping = {
            'null': 'void',
            'boolean': 'boolean',
            'string': 'String',
            'integer': 'int',
            'number': 'double',
            'int8': 'byte',
            'uint8': 'short',
            'int16': 'short',
            'uint16': 'int',
            'int32': 'int',
            'uint32': 'long',
            'int64': 'long',
            'uint64': 'BigInteger',  # Use BigInteger for uint64 to handle full range in JSON
            'int128': 'BigInteger',
            'uint128': 'BigInteger',
            'float8': 'float',
            'float': 'float',
            'double': 'double',
            'binary32': 'float',
            'binary64': 'double',
            'decimal': 'BigDecimal',
            'binary': 'byte[]',
            'date': 'LocalDate',
            'time': 'LocalTime',
            'datetime': 'Instant',
            'timestamp': 'Instant',
            'duration': 'Duration',
            'uuid': 'UUID',
            'uri': 'URI',
            'jsonpointer': 'String',
            'any': 'Object'
        }
        if '.' in structure_type:
            type_name = structure_type.split('.')[-1]
            package_name = '.'.join(structure_type.split('.')[:-1]).lower()
            structure_type = self.qualified_name(package_name, type_name)
        if structure_type in self.generated_types_structure_namespace:
            kind = self.generated_types_structure_namespace[structure_type]
            qualified_class_name = self.qualified_name(self.base_package, structure_type)
            return StructureToJava.JavaType(qualified_class_name, is_class=kind=="class", is_enum=kind=="enum")
        else:
            return StructureToJava.JavaType(required_mapping.get(structure_type, structure_type) if not is_optional else optional_mapping.get(structure_type, structure_type))
    
    def is_java_primitive(self, java_type: JavaType) -> bool:
        """Checks if a Java type is a primitive type"""
        return java_type.type_name in [
            'void', 'boolean', 'int', 'long', 'float', 'double', 'byte', 'short', 'byte[]', 'String',
            'Boolean', 'Integer', 'Long', 'Float', 'Double', 'Short', 'Byte', 'Void']
        
    def is_java_optional_type(self, java_type: JavaType) -> bool:
        """Checks if a Java type is an optional type"""
        return java_type.type_name in ['Void', 'Boolean', 'Integer', 'Long', 'Float', 'Double', 'Short', 'Byte']
        
    def is_java_numeric_type(self, java_type: JavaType) -> bool:
        """Checks if a Java type is a numeric type"""
        return java_type.type_name in ['int', 'long', 'float', 'double', 'short', 'byte', 'Integer', 'Long', 'Float', 'Double', 'Short', 'Byte', 'BigInteger', 'BigDecimal']
    
    def is_java_integer_type(self, java_type: JavaType) -> bool:
        """Checks if a Java type is an integer type"""
        return java_type.type_name in ['int', 'long', 'short', 'byte', 'Integer', 'Long', 'Short', 'Byte', 'BigInteger']

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """ Resolves a $ref to the actual schema definition """
        # Check if it's an absolute URI reference (schema with $id)
        if not ref.startswith('#/'):
            # Try to resolve from schema registry
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None
        
        # Handle fragment-only references (internal to document)
        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc
        
        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]
        
        return schema

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """ Recursively registers schemas with $id keywords """
        if not isinstance(schema, dict):
            return
        
        # Register this schema if it has an $id
        if '$id' in schema:
            schema_id = schema['$id']
            # Handle relative URIs
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id  # Update base URI for nested schemas
        
        # Recursively process definitions
        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)
        
        # Recursively process properties
        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)
        
        # Recursively process items, values, etc.
        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def convert_structure_type_to_java(self, class_name: str, field_name: str, structure_type: Union[str, Dict, List], parent_package: str, nullable: bool = False) -> JavaType:
        """Converts JSON Structure type to Java type"""
        if isinstance(structure_type, str):
            return self.map_primitive_to_java(structure_type, nullable)
        elif isinstance(structure_type, list):
            # Handle type unions
            non_null_types = [t for t in structure_type if t != 'null']
            if len(non_null_types) == 1:
                if isinstance(non_null_types[0], str):
                    return self.map_primitive_to_java(non_null_types[0], True)
                else:
                    return self.convert_structure_type_to_java(class_name, field_name, non_null_types[0], parent_package)
            else:
                # Multiple non-null types - generate union class
                if self.jackson_annotations:
                    return StructureToJava.JavaType(self.generate_embedded_union_class_jackson(class_name, field_name, non_null_types, parent_package, write_file=True), is_class=True)
                else:
                    types: List[StructureToJava.JavaType] = [self.convert_structure_type_to_java(
                        class_name, field_name, t, parent_package) for t in non_null_types]
                    return StructureToJava.JavaType('Object', types)
        elif isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                ref_schema = self.resolve_ref(structure_type['$ref'], self.schema_doc if isinstance(self.schema_doc, dict) else None)
                if ref_schema:
                    ref_path = structure_type['$ref'].split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_package
                    return self.generate_class_or_enum(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                return StructureToJava.JavaType('Object')
            
            # Handle enum keyword
            if 'enum' in structure_type:
                return self.generate_enum(structure_type, field_name, parent_package, write_file=True)
            
            # Handle type keyword
            if 'type' not in structure_type:
                return StructureToJava.JavaType('Object')
            
            struct_type = structure_type['type']
            
            # Handle complex types
            if struct_type == 'object':
                return self.generate_class(structure_type, parent_package, write_file=True)
            elif struct_type == 'array':
                item_type = self.convert_structure_type_to_java(class_name, field_name, structure_type.get('items', {'type': 'any'}), parent_package, nullable=True).type_name
                return StructureToJava.JavaType(f"List<{item_type}>")
            elif struct_type == 'set':
                item_type = self.convert_structure_type_to_java(class_name, field_name, structure_type.get('items', {'type': 'any'}), parent_package, nullable=True).type_name
                return StructureToJava.JavaType(f"Set<{item_type}>")
            elif struct_type == 'map':
                values_type = self.convert_structure_type_to_java(class_name, field_name, structure_type.get('values', {'type': 'any'}), parent_package, nullable=True).type_name
                return StructureToJava.JavaType(f"Map<String,{values_type}>")
            elif struct_type == 'choice':
                return self.generate_choice(structure_type, parent_package, write_file=True)
            elif struct_type == 'tuple':
                return self.generate_tuple(structure_type, parent_package, write_file=True)
            else:
                return self.convert_structure_type_to_java(class_name, field_name, struct_type, parent_package)
        return StructureToJava.JavaType('Object')

    def generate_class_or_enum(self, structure_schema: Dict, parent_package: str, write_file: bool = True, explicit_name: str = '') -> JavaType:
        """ Generates a Java class or enum from a JSON Structure schema """
        struct_type = structure_schema.get('type', 'object')
        if 'enum' in structure_schema:
            return self.generate_enum(structure_schema, explicit_name or structure_schema.get('name', 'UnnamedEnum'), parent_package, write_file)
        elif struct_type == 'object':
            return self.generate_class(structure_schema, parent_package, write_file, explicit_name=explicit_name)
        elif struct_type == 'choice':
            return self.generate_choice(structure_schema, parent_package, write_file, explicit_name=explicit_name)
        elif struct_type == 'tuple':
            return self.generate_tuple(structure_schema, parent_package, write_file, explicit_name=explicit_name)
        return StructureToJava.JavaType('Object')

    def generate_class(self, structure_schema: Dict, parent_package: str, write_file: bool, explicit_name: str = '') -> JavaType:
        """ Generates a Java class from a JSON Structure object schema """
        
        # Get name and namespace
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedClass'))
        schema_namespace = structure_schema.get('namespace', parent_package)
        if not 'namespace' in structure_schema:
            structure_schema['namespace'] = schema_namespace
        package = self.join_packages(self.base_package, schema_namespace).replace('.', '/').lower()
        package = package.replace('.', '/').lower()
        package = self.safe_package(package)
        class_name = self.safe_identifier(class_name)
        namespace_qualified_name = self.qualified_name(schema_namespace, explicit_name or structure_schema.get('name', 'UnnamedClass'))
        qualified_class_name = self.qualified_name(package.replace('/', '.'), class_name)
        if namespace_qualified_name in self.generated_types_structure_namespace:
            return StructureToJava.JavaType(qualified_class_name, is_class=True)
        self.generated_types_structure_namespace[namespace_qualified_name] = "class"
        self.generated_types_java_package[qualified_class_name] = "class"
        
        # Check if this is an abstract type
        is_abstract = structure_schema.get('abstract', False)
        deprecated = structure_schema.get('deprecated', False)
        
        # Generate documentation
        doc = structure_schema.get('description', structure_schema.get('doc', class_name))
        
        # Handle inheritance
        base_class = None
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            base_schema = self.resolve_ref(base_ref, self.schema_doc if isinstance(self.schema_doc, dict) else None)
            if base_schema:
                ref_path = base_ref.split('/')
                base_name = ref_path[-1]
                ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_package
                base_type = self.generate_class(base_schema, ref_namespace, write_file=True, explicit_name=base_name)
                base_class = base_type.type_name
        
        # Generate field information for template
        fields = []
        for prop_name, prop_schema in structure_schema.get('properties', {}).items():
            field_info = self.generate_field_info(class_name, prop_name, prop_schema, schema_namespace)
            fields.append(field_info)
        
        # Use template for class generation
        class_definition = process_template(
            "structuretojava/class_core.jinja",
            class_name=class_name,
            docstring=doc,
            is_abstract=is_abstract,
            deprecated=deprecated,
            base_class=base_class,
            fields=fields,
            jackson_annotation=self.jackson_annotations
        )
        
        # Generate Equals and GetHashCode using template
        properties = structure_schema.get('properties', {})
        non_const_properties = {k: v for k, v in properties.items() if 'const' not in v}
        field_names = []
        for prop_name in non_const_properties.keys():
            field_name_java = pascal(prop_name) if self.pascal_properties else prop_name
            safe_field_name = self.safe_identifier(field_name_java, class_name)
            field_names.append(safe_field_name)
        
        equals_hashcode = process_template(
            "structuretojava/equals_hashcode.jinja",
            class_name=class_name,
            fields=field_names,
            field_count=len(field_names)
        )
        
        class_definition = class_definition.rstrip() + equals_hashcode + "\n}\n"

        if write_file:
            self.write_to_file(package, class_name, class_definition)
        return StructureToJava.JavaType(qualified_class_name, is_class=True)

    def generate_field_info(self, class_name: str, prop_name: str, prop_schema: Dict, parent_package: str) -> Dict:
        """ Generates field information for template """
        field_name_java = pascal(prop_name) if self.pascal_properties else prop_name
        field_type = self.convert_structure_type_to_java(class_name, prop_name, prop_schema, parent_package)
        safe_field_name = self.safe_identifier(field_name_java, class_name)
        
        # Generate documentation
        doc = prop_schema.get('description', prop_schema.get('doc', field_name_java))
        
        # Check if this is a const field
        is_const = 'const' in prop_schema
        const_value = None
        if is_const:
            const_val = prop_schema['const']
            prop_type = field_type.type_name
            if prop_type.endswith('?'):
                prop_type = prop_type[:-1]
            const_value = self.format_const_value(const_val, prop_type)
        
        return {
            'name': safe_field_name,
            'original_name': prop_name,
            'type': field_type.type_name,
            'docstring': doc,
            'is_const': is_const,
            'const_value': const_value
        }
    
    def generate_property(self, class_name: str, field: Tuple[str, Dict], parent_package: str) -> str:
        """ Generates a Java property definition (legacy method for compatibility) """
        field_name, field_schema = field
        field_info = self.generate_field_info(class_name, field_name, field_schema, parent_package)
        
        property_def = f"{INDENT}/** {field_info['docstring']} */\n"
        
        if self.jackson_annotations and field_info['original_name'] != field_info['name']:
            property_def += f'{INDENT}@JsonProperty("{field_info["original_name"]}")\n'
        
        if field_info['is_const']:
            property_def += f"{INDENT}public static final {field_info['type']} {field_info['name']} = {field_info['const_value']};\n"
        else:
            property_def += f"{INDENT}private {field_info['type']} {field_info['name']};\n"
            property_def += f"{INDENT}public {field_info['type']} get{pascal(field_info['name'])}() {{ return {field_info['name']}; }}\n"
            property_def += f"{INDENT}public void set{pascal(field_info['name'])}({field_info['type']} {field_info['name']}) {{ this.{field_info['name']} = {field_info['name']}; }}\n"
        
        return property_def

    def format_const_value(self, value: Any, java_type: str) -> str:
        """ Formats a constant value for Java """
        if value is None:
            return "null"
        elif isinstance(value, bool):
            return "true" if value else "false"
        elif isinstance(value, str):
            return f'"{value}"'
        elif isinstance(value, (int, float)):
            if java_type in ['float', 'Float']:
                return f"{value}f"
            elif java_type in ['long', 'Long']:
                return f"{value}L"
            elif java_type in ['double', 'Double']:
                return f"{value}d"
            return str(value)
        return f"/* unsupported const value */"

    def generate_enum(self, structure_schema: Dict, field_name: str, parent_package: str, write_file: bool) -> JavaType:
        """ Generates a Java enum from JSON Structure enum schema """
        
        # Determine enum name
        enum_name = pascal(structure_schema.get('name', field_name + 'Enum'))
        schema_namespace = structure_schema.get('namespace', parent_package)
        package = self.join_packages(self.base_package, schema_namespace).replace('.', '/').lower()       
        enum_name = self.safe_identifier(enum_name)
        type_name = self.qualified_name(package.replace('/', '.'), enum_name)
        namespace_qualified_name = self.qualified_name(schema_namespace, structure_schema.get('name', field_name + 'Enum'))
        self.generated_types_structure_namespace[namespace_qualified_name] = "enum"
        self.generated_types_java_package[type_name] = "enum"
        
        # Get enum values
        symbols = structure_schema.get('enum', [])
        if not symbols:
            return StructureToJava.JavaType('Object')
        
        # Generate documentation
        doc = structure_schema.get('description', structure_schema.get('doc', enum_name))
        deprecated = structure_schema.get('deprecated', False)
        
        # Determine base type
        base_type = structure_schema.get('type', 'string')
        numeric_types = ['int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32', 'int64', 'uint64']
        is_numeric = base_type in numeric_types
        
        if is_numeric:
            java_base_type = self.map_primitive_to_java(base_type, False).type_name
            symbol_list = [{'name': f"VALUE_{value}".upper(), 'value': value} for value in symbols]
            enum_definition = process_template(
                "structuretojava/enum_core.jinja",
                class_name=enum_name,
                docstring=doc,
                deprecated=deprecated,
                is_numeric=True,
                numeric_type=java_base_type,
                symbols=symbol_list
            )
        else:
            # String enum
            safe_symbols = [self.safe_identifier(pascal(str(symbol).replace('-', '_').replace(' ', '_'))) for symbol in symbols]
            enum_definition = process_template(
                "structuretojava/enum_core.jinja",
                class_name=enum_name,
                docstring=doc,
                deprecated=deprecated,
                is_numeric=False,
                symbols=safe_symbols
            )
        
        if write_file:
            self.write_to_file(package, enum_name, enum_definition)
        return StructureToJava.JavaType(type_name, is_enum=True)

    def generate_choice(self, structure_schema: Dict, parent_package: str, write_file: bool, explicit_name: str = '') -> JavaType:
        """ Generates a choice (discriminated union) type """
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedChoice'))
        schema_namespace = structure_schema.get('namespace', parent_package)
        package = self.join_packages(self.base_package, schema_namespace).replace('.', '/').lower()
        package = self.safe_package(package)
        class_name = self.safe_identifier(class_name)
        namespace_qualified_name = self.qualified_name(schema_namespace, explicit_name or structure_schema.get('name', 'UnnamedChoice'))
        qualified_class_name = self.qualified_name(package.replace('/', '.'), class_name)
        
        if namespace_qualified_name in self.generated_types_structure_namespace:
            return StructureToJava.JavaType(qualified_class_name, is_class=True)
        
        self.generated_types_structure_namespace[namespace_qualified_name] = "class"
        self.generated_types_java_package[qualified_class_name] = "class"
        
        # Get choice definitions
        choices_dict = structure_schema.get('choices', {})
        doc = structure_schema.get('description', structure_schema.get('doc', class_name))
        deprecated = structure_schema.get('deprecated', False)
        
        # Build choice information for template
        choices = []
        for choice_name, choice_schema in choices_dict.items():
            choice_type_name = pascal(choice_name)
            value_type = self.convert_structure_type_to_java(class_name, choice_name, choice_schema, schema_namespace)
            choices.append({
                'name': choice_name,
                'type': choice_type_name,
                'value_type': value_type.type_name,
                'docstring': choice_schema.get('description', choice_schema.get('doc', f'{choice_name} variant'))
            })
        
        # Use template for choice generation
        choice_definition = process_template(
            "structuretojava/choice_core.jinja",
            class_name=class_name,
            docstring=doc,
            deprecated=deprecated,
            choices=choices,
            jackson_annotation=self.jackson_annotations
        )
        
        if write_file:
            self.write_to_file(package, class_name, choice_definition)
        
        return StructureToJava.JavaType(qualified_class_name, is_class=True)

    def generate_tuple(self, structure_schema: Dict, parent_package: str, write_file: bool, explicit_name: str = '') -> JavaType:
        """ Generates a tuple type - Per JSON Structure spec, tuples serialize as JSON arrays """
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedTuple'))
        schema_namespace = structure_schema.get('namespace', parent_package)
        package = self.join_packages(self.base_package, schema_namespace).replace('.', '/').lower()
        package = self.safe_package(package)
        class_name = self.safe_identifier(class_name)
        namespace_qualified_name = self.qualified_name(schema_namespace, explicit_name or structure_schema.get('name', 'UnnamedTuple'))
        qualified_class_name = self.qualified_name(package.replace('/', '.'), class_name)
        
        if namespace_qualified_name in self.generated_types_structure_namespace:
            return StructureToJava.JavaType(qualified_class_name, is_class=True)
        
        self.generated_types_structure_namespace[namespace_qualified_name] = "class"
        self.generated_types_java_package[qualified_class_name] = "class"
        
        # Get tuple order and properties
        properties = structure_schema.get('properties', {})
        tuple_order = structure_schema.get('tuple', [])
        doc = structure_schema.get('description', structure_schema.get('doc', class_name))
        deprecated = structure_schema.get('deprecated', False)
        
        # Build tuple element information in correct order
        elements = []
        for prop_name in tuple_order:
            if prop_name in properties:
                prop_schema = properties[prop_name]
                prop_type = self.convert_structure_type_to_java(class_name, prop_name, prop_schema, schema_namespace)
                field_name = pascal(prop_name) if self.pascal_properties else prop_name
                safe_field_name = self.safe_identifier(field_name, class_name)
                elements.append({
                    'name': safe_field_name,
                    'type': prop_type.type_name,
                    'docstring': prop_schema.get('description', prop_schema.get('doc', prop_name))
                })
        
        # Use template for tuple generation
        tuple_definition = process_template(
            "structuretojava/tuple_core.jinja",
            class_name=class_name,
            docstring=doc,
            deprecated=deprecated,
            elements=elements,
            jackson_annotation=self.jackson_annotations
        )
        
        if write_file:
            self.write_to_file(package, class_name, tuple_definition)
        
        return StructureToJava.JavaType(qualified_class_name, is_class=True)

    def generate_embedded_union_class_jackson(self, class_name: str, field_name: str, structure_types: List, parent_package: str, write_file: bool) -> str:
        """ Generates an embedded Union Class for Java using Jackson """
        # Simplified implementation - just return Object for unions
        return 'Object'

    def generate_equals_and_gethashcode(self, structure_schema: Dict, class_name: str, parent_namespace: str) -> str:
        """ Generates Equals and GetHashCode methods """
        code = "\n"
        properties = structure_schema.get('properties', {})
        
        # Filter out const properties
        non_const_properties = {k: v for k, v in properties.items() if 'const' not in v}
        
        if not non_const_properties:
            # Empty class
            code += f"{INDENT}@Override\n{INDENT}public boolean equals(Object obj) {{\n"
            code += f"{INDENT*2}return obj instanceof {class_name};\n"
            code += f"{INDENT}}}\n\n"
            code += f"{INDENT}@Override\n{INDENT}public int hashCode() {{\n"
            code += f"{INDENT*2}return 0;\n"
            code += f"{INDENT}}}\n"
            return code
        
        # Generate equals
        code += f"{INDENT}@Override\n{INDENT}public boolean equals(Object obj) {{\n"
        code += f"{INDENT*2}if (this == obj) return true;\n"
        code += f"{INDENT*2}if (!(obj instanceof {class_name})) return false;\n"
        code += f"{INDENT*2}{class_name} other = ({class_name}) obj;\n"
        
        equality_checks = []
        for prop_name in non_const_properties.keys():
            field_name_java = pascal(prop_name) if self.pascal_properties else prop_name
            safe_field_name = self.safe_identifier(field_name_java, class_name)
            equality_checks.append(f"Objects.equals(this.{safe_field_name}, other.{safe_field_name})")
        
        if len(equality_checks) == 1:
            code += f"{INDENT*2}return {equality_checks[0]};\n"
        else:
            code += f"{INDENT*2}return " + f"\n{INDENT*3}&& ".join(equality_checks) + ";\n"
        
        code += f"{INDENT}}}\n\n"
        
        # Generate hashCode
        code += f"{INDENT}@Override\n{INDENT}public int hashCode() {{\n"
        
        hash_fields = []
        for prop_name in non_const_properties.keys():
            field_name_java = pascal(prop_name) if self.pascal_properties else prop_name
            safe_field_name = self.safe_identifier(field_name_java, class_name)
            hash_fields.append(safe_field_name)
        
        if len(hash_fields) <= 8:
            code += f"{INDENT*2}return Objects.hash({', '.join(hash_fields)});\n"
        else:
            code += f"{INDENT*2}int result = Objects.hash({', '.join(hash_fields[:8])});\n"
            for i in range(8, len(hash_fields)):
                code += f"{INDENT*2}result = 31 * result + Objects.hashCode({hash_fields[i]});\n"
            code += f"{INDENT*2}return result;\n"
        
        code += f"{INDENT}}}\n"
        
        return code

    def write_to_file(self, package: str, name: str, definition: str):
        """ Writes a Java class or enum to a file """
        package = package.lower()
        package = self.safe_package(package)
        directory_path = os.path.join(
            self.output_dir, package.replace('.', os.sep).replace('/', os.sep))
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)
        file_path = os.path.join(directory_path, f"{name}.java")

        with open(file_path, 'w', encoding='utf-8') as file:
            if package:
                file.write(f"package {package.replace('/', '.')};\n\n")
                if "List<" in definition:
                    file.write("import java.util.List;\n")
                if "Set<" in definition:
                    file.write("import java.util.Set;\n")
                if "Map<" in definition:
                    file.write("import java.util.Map;\n")
                if "BigDecimal" in definition:
                    file.write("import java.math.BigDecimal;\n")
                if "BigInteger" in definition:
                    file.write("import java.math.BigInteger;\n")
                if "LocalDate" in definition:
                    file.write("import java.time.LocalDate;\n")
                if "LocalTime" in definition:
                    file.write("import java.time.LocalTime;\n")
                if "Instant" in definition:
                    file.write("import java.time.Instant;\n")
                if "LocalDateTime" in definition:
                    file.write("import java.time.LocalDateTime;\n")
                if "UUID" in definition:
                    file.write("import java.util.UUID;\n")
                if "Duration" in definition:
                    file.write("import java.time.Duration;\n")
                if "URI" in definition:
                    file.write("import java.net.URI;\n")
                if "Objects" in definition:
                    file.write("import java.util.Objects;\n")
                    
            if self.jackson_annotations:
                if 'JsonProperty' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonProperty;\n")
                if 'JsonNode' in definition:
                    file.write("import com.fasterxml.jackson.databind.JsonNode;\n")
                if 'ObjectMapper' in definition:
                    file.write("import com.fasterxml.jackson.databind.ObjectMapper;\n")
                if 'JsonSerialize' in definition:
                    file.write("import com.fasterxml.jackson.databind.annotation.JsonSerialize;\n")
                if 'JsonDeserialize' in definition:
                    file.write("import com.fasterxml.jackson.databind.annotation.JsonDeserialize;\n")               
                if 'JsonSerializer' in definition:
                    file.write("import com.fasterxml.jackson.databind.JsonSerializer;\n") 
                if 'SerializerProvider' in definition:
                    file.write("import com.fasterxml.jackson.databind.SerializerProvider;\n")
                if 'JsonDeserializer' in definition:
                    file.write("import com.fasterxml.jackson.databind.JsonDeserializer;\n")
                if 'DeserializationContext' in definition:
                    file.write("import com.fasterxml.jackson.databind.DeserializationContext;\n")
                if 'JsonParser' in definition:
                    file.write("import com.fasterxml.jackson.core.JsonParser;\n")
                if 'JsonIgnore' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonIgnore;\n")
                if 'JsonProcessingException' in definition:
                    file.write("import com.fasterxml.jackson.core.JsonProcessingException;\n")
                if 'JsonGenerator' in definition:
                    file.write("import com.fasterxml.jackson.core.JsonGenerator;\n")
                if 'TypeReference' in definition:
                    file.write("import com.fasterxml.jackson.core.type.TypeReference;\n")
                if 'JsonFormat' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonFormat;\n")
                if 'JsonCreator' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonCreator;\n")
                if 'JsonValue' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonValue;\n")
                if 'JsonTypeInfo' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonTypeInfo;\n")
                if 'JsonSubTypes' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonSubTypes;\n")
                    file.write("import com.fasterxml.jackson.core.JsonGenerator;\n")
                if 'TypeReference' in definition:
                    file.write("import com.fasterxml.jackson.core.type.TypeReference;\n")
            file.write("\n")
            file.write(definition)

    def convert_schema(self, schema: JsonNode, output_dir: str):
        """Converts JSON Structure schema to Java"""
        if not isinstance(schema, list):
            schema = [schema]
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        pom_path = os.path.join(output_dir, "pom.xml")
        if not os.path.exists(pom_path):
            package_elements = self.base_package.split('.') if self.base_package else ["com", "example"]
            groupid = '.'.join(package_elements[:-1]) if len(package_elements) > 1 else package_elements[0]
            artifactid = package_elements[-1]
            pom_content = process_template(
                "structuretojava/pom.xml.jinja",
                groupid=groupid,
                artifactid=artifactid,
                jackson_version=JACKSON_VERSION
            )
            with open(pom_path, 'w', encoding='utf-8') as file:
                file.write(pom_content)
        output_dir = os.path.join(
            output_dir, "src/main/java".replace('/', os.sep))
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        self.output_dir = output_dir
        
        # Register all schemas with $id keywords
        for structure_schema in (x for x in schema if isinstance(x, dict)):
            self.schema_doc = structure_schema
            self.register_schema_ids(structure_schema)
        
        # Generate classes
        for structure_schema in (x for x in schema if isinstance(x, dict)):
            self.schema_doc = structure_schema
            if 'definitions' in structure_schema:
                self.process_definitions(structure_schema['definitions'], '')
            if 'type' in structure_schema or 'enum' in structure_schema:
                self.generate_class_or_enum(structure_schema, '')

    def process_definitions(self, definitions: Dict, namespace_path: str) -> None:
        """ Processes the definitions section recursively """
        for name, definition in definitions.items():
            if isinstance(definition, dict):
                if 'type' in definition or 'enum' in definition:
                    # This is a type definition
                    current_namespace = namespace_path
                    self.generate_class_or_enum(definition, current_namespace, write_file=True, explicit_name=name)
                else:
                    # This might be a nested namespace
                    new_namespace = f"{namespace_path}.{name}" if namespace_path else name
                    self.process_definitions(definition, new_namespace)

    def convert(self, structure_schema_path: str, output_dir: str):
        """Converts JSON Structure schema to Java"""
        with open(structure_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)


def convert_structure_to_java(structure_schema_path, java_file_path, package_name='', pascal_properties=False, jackson_annotation=True):
    """
    Converts JSON Structure schema to Java classes

    Args:
        structure_schema_path: JSON Structure input schema path  
        java_file_path: Output Java directory path
        package_name: Base package name
        pascal_properties: Use PascalCase for properties
        jackson_annotation: Add Jackson annotations (always True for Structure)
    """
    if not package_name:
        package_name = os.path.splitext(os.path.basename(java_file_path))[0].replace('-', '_').lower()
    structuretojava = StructureToJava()
    structuretojava.base_package = package_name
    structuretojava.pascal_properties = pascal_properties
    structuretojava.jackson_annotations = jackson_annotation
    structuretojava.convert(structure_schema_path, java_file_path)


def convert_structure_schema_to_java(structure_schema: JsonNode, output_dir: str, package_name='', pascal_properties=False, jackson_annotation=True):
    """
    Converts JSON Structure schema to Java classes

    Args:
        structure_schema: JSON Structure schema as a dictionary or list of dictionaries
        output_dir: Output directory path
        package_name: Base package name
        pascal_properties: Use PascalCase for properties
        jackson_annotation: Add Jackson annotations (always True for Structure)
    """
    structuretojava = StructureToJava()
    structuretojava.base_package = package_name
    structuretojava.pascal_properties = pascal_properties
    structuretojava.jackson_annotations = jackson_annotation
    structuretojava.convert_schema(structure_schema, output_dir)



import json
import sys
from typing import Dict, List, cast, Optional
from datapackage import Package, Resource

from avrotize.common import get_longest_namespace_prefix


JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | bool | int | None

class AvroToDataPackageConverter:
    """Class to convert Avro schema to Data Package."""

    def __init__(self: 'AvroToDataPackageConverter'):
        self.named_type_cache: Dict[str, JsonNode] = {}

    def get_fullname(self, namespace: str, name: str) -> str:
        """Get the full name of a record type."""
        return f"{namespace}.{name}" if namespace else name

    def convert_avro_to_datapackage(self, avro_schema_path: str, avro_record_type: Optional[str], datapackage_path: str):
        """Convert an Avro schema to a Data Package."""
        with open(avro_schema_path, "r", encoding="utf-8") as f:
            schema_json = f.read()

        # Parse the schema as a JSON object
        schema = json.loads(schema_json)
        self.cache_named_types(schema)

        if isinstance(schema, list):
            if avro_record_type:
                schema = next(
                    (x for x in schema if x["name"] == avro_record_type or x["namespace"] + "." + x["name"] == avro_record_type), None)
                if schema is None:
                    print(f"No top-level record type {avro_record_type} found in the Avro schema")
                    sys.exit(1)
            schemas_to_convert = schema
        elif isinstance(schema, dict):
            schemas_to_convert = [schema]
        else:
            print("Expected a single Avro schema as a JSON object, or a list of schema records")
            sys.exit(1)

        longest_namespace_prefix = get_longest_namespace_prefix(schema)
        self.create_datapackage_for_schemas(schemas_to_convert, datapackage_path, longest_namespace_prefix)

    def create_datapackage_for_schemas(self, schemas: List[Dict[str, JsonNode]], datapackage_path: str, namespace_prefix: str):
        """Create a Data Package for given schemas."""
        package = Package()
        data_package_resources = []

        for schema in schemas:
            name = str(schema["name"])
            namespace = str(schema.get("namespace", ""))
            if namespace.startswith(namespace_prefix):
                namespace = namespace[len(namespace_prefix):].strip(".")
            table_name = f"{namespace}_{name}" if namespace else name
            fields = cast(List[Dict[str, JsonNode]], schema["fields"])

            # Create the Data Package schema
            resource_schema: Dict[str, List[JsonNode]] = {
                "fields": []
            }

            for field in fields:
                column_name = field["name"]
                column_type = self.convert_avro_type_to_datapackage_type(field["type"])
                field_schema = {"name": column_name, "type": column_type}
                if "doc" in field:
                    field_schema["description"] = field["doc"]
                resource_schema["fields"].append(field_schema)

            resource = {
                "name": table_name,
                "schema": resource_schema
            }
            data_package_resources.append(resource)

        # Add resources to the Data Package
        for resource in data_package_resources:
            package.add_resource(resource)

        # Save the Data Package
        package.descriptor["name"] = namespace_prefix
        package.commit()

        with open(datapackage_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(package.descriptor, indent=2))

    def convert_avro_type_to_datapackage_type(self, avro_type: JsonNode) -> str:
        """Convert an Avro type to a Data Package type."""
        if isinstance(avro_type, list):
            item_count = len(avro_type)
            if item_count == 1:
                return self.convert_avro_type_to_datapackage_type(avro_type[0])
            elif item_count == 2:
                first = avro_type[0]
                second = avro_type[1]
                if isinstance(first, str) and first == "null":
                    return self.convert_avro_type_to_datapackage_type(second)
                elif isinstance(second, str) and second == "null":
                    return self.convert_avro_type_to_datapackage_type(first)
            print(f"WARNING: Complex union types are not fully supported: {avro_type}")
            return "string"
        elif isinstance(avro_type, dict):
            type_name = avro_type.get("type")
            if type_name == "array":
                return "array"
            elif type_name == "map":
                return "object"
            elif type_name == "record":
                return "object"
            elif type_name == "enum":
                return "string"
            elif type_name == "fixed":
                return "string"
            elif type_name == "string":
                return "string"
            elif type_name == "bytes":
                return "string"
            elif type_name == "long":
                return "integer"
            elif type_name == "int":
                return "integer"
            elif type_name == "float":
                return "number"
            elif type_name == "double":
                return "number"
            elif type_name == "boolean":
                return "boolean"
            else:
                return "string"
        elif isinstance(avro_type, str):
            if avro_type in self.named_type_cache:
                return self.convert_avro_type_to_datapackage_type(self.named_type_cache[avro_type])
            return self.map_scalar_type(avro_type)

        return "string"

    def cache_named_types(self, avro_type: JsonNode):
        """Add an encountered type to the list of types."""
        if isinstance(avro_type, list):
            for item in avro_type:
                self.cache_named_types(item)
        if isinstance(avro_type, dict) and avro_type.get("name"):
            self.named_type_cache[self.get_fullname(str(avro_type.get(
                "namespace")), str(avro_type.get("name")))] = avro_type
            if "fields" in avro_type:
                for field in cast(List[Dict[str,JsonNode]],avro_type.get("fields")):
                    if "type" in field:
                        self.cache_named_types(field.get("type"))

    def map_scalar_type(self, type_name: str) -> str:
        """Map an Avro scalar type to a Data Package scalar type."""
        scalar_type_mapping = {
            "null": "string",
            "int": "integer",
            "long": "integer",
            "float": "number",
            "double": "number",
            "boolean": "boolean",
            "bytes": "string",
            "string": "string"
        }
        return scalar_type_mapping.get(type_name, "string")

def convert_avro_to_datapackage(avro_schema_path: str, avro_record_type: Optional[str], datapackage_path: str):
    """Convert an Avro schema to a Data Package."""
    converter = AvroToDataPackageConverter()
    converter.convert_avro_to_datapackage(avro_schema_path, avro_record_type, datapackage_path)

# Example usage:
# convert_avro_to_datapackage("schema.avsc", "MyRecord", "datapackage.json")



"""
CDDL (Concise Data Definition Language) to JSON Structure converter.

CDDL is defined in RFC 8610 and is a schema language primarily used for
expressing CBOR and JSON data structures.

RFC 8610 Compliance Summary:
============================

Fully Implemented (RFC 8610 Sections 2-3):
- Primitive types: uint, nint, int, float16/32/64, tstr, bstr, bool, nil, any
- Maps/objects with member keys and values
- Arrays with homogeneous and heterogeneous (tuple) items
- Choice/union types (type1 / type2)
- Occurrence indicators: ? (optional), * (zero-or-more), + (one-or-more)
- Literal values (strings, integers, floats, booleans)
- Range constraints (min..max, min...max)
- Type references and nested types
- Groups and group composition
- CBOR tags (#6.n(type)) - underlying type extracted
- Choice from groups (&group)
- Comments (handled by parser)
- Generic type parameters (<T>): Parameterized type definitions with instantiation
- Unwrap operator (~): Extracting group/array content

Control Operators (RFC 8610 Section 3.8):
- .size: Maps to minLength/maxLength for strings, minItems/maxItems for arrays
- .regexp: Maps to pattern validation (ECMAScript regex)
- .default: Maps to default value
- .lt, .le, .gt, .ge: Maps to minimum/maximum/exclusiveMinimum/exclusiveMaximum
- .eq: Maps to const value
- .bits, .cbor, .cborseq: Base type extracted (CBOR-specific, no JSON equivalent)
- .within, .and: Type intersection (limited support, base type used)

Not Implemented:
- CDDL sockets ($name, $$name): Extensibility points (parser limitation)
- .ne operator: No direct JSON Structure equivalent

Notes on Type Mapping:
- CBOR-specific types (bstr, tstr) map to JSON equivalents (bytes, string)
- CBOR tags are unwrapped to their underlying types
- Integer types (uint, nint, int) all map to int64 for JSON compatibility
- Float types map to float (float16/32) or double (float64)
"""
# pylint: disable=too-many-lines,too-many-instance-attributes,too-many-arguments
# pylint: disable=too-many-positional-arguments,too-many-locals,too-many-branches
# pylint: disable=too-many-statements,too-many-return-statements,too-many-nested-blocks
# pylint: disable=no-else-return

import json
import logging
from typing import Any, Dict, List, Optional, Set, Tuple

from cddlparser import parse as cddl_parse
from cddlparser.ast import (
    Rule, Type, Typename, Map, Array, Value, Range, Operator, Tag,
    ChoiceFrom, CDDLNode
)
from json_structure import SchemaValidator, ValidationError, ValidationSeverity

from avrotize.common import avro_name


# Configure module logger
logger = logging.getLogger(__name__)


# Default namespace for JSON Structure schema
DEFAULT_NAMESPACE = 'example.com'

# Maximum recursion depth for type conversion (prevents stack overflow)
MAX_CONVERSION_DEPTH = 100

# Type aliases for better type safety
JsonStructureType = Dict[str, Any]
AstNode = Any  # Union of all cddlparser AST node types
CommentList = List[Any]


class CddlConversionError(Exception):
    """
    Exception raised when CDDL to JSON Structure conversion fails.

    Attributes:
        message: Human-readable error description
        context: Optional context about where the error occurred
        cause: Optional underlying exception that caused this error
    """

    def __init__(self, message: str, context: Optional[str] = None,
                 cause: Optional[Exception] = None) -> None:
        self.message = message
        self.context = context
        self.cause = cause
        full_message = message
        if context:
            full_message = f"{message} (context: {context})"
        super().__init__(full_message)


class CddlCycleError(CddlConversionError):
    """
    Exception raised when a circular type reference is detected.

    Attributes:
        cycle_path: List of type names forming the cycle
    """

    def __init__(self, cycle_path: List[str]) -> None:
        self.cycle_path = cycle_path
        cycle_str = ' -> '.join(cycle_path)
        super().__init__(f"Circular type reference detected: {cycle_str}")


# CDDL primitive types to JSON Structure type mapping
CDDL_PRIMITIVE_TYPES: Dict[str, Dict[str, Any]] = {
    # Integer types
    'uint': {'type': 'int64'},      # Unsigned integer (CBOR major type 0)
    'nint': {'type': 'int64'},      # Negative integer (CBOR major type 1)
    'int': {'type': 'int64'},       # Signed integer (uint or nint)
    'integer': {'type': 'int64'},   # Alias for int

    # Floating-point types
    'float16': {'type': 'float'},   # Half precision
    'float32': {'type': 'float'},   # Single precision
    'float64': {'type': 'double'},  # Double precision
    'float': {'type': 'double'},    # Generic float (maps to double)

    # String types
    'bstr': {'type': 'binary'},      # Byte string (base64 encoded)
    'bytes': {'type': 'binary'},     # Alias for bstr
    'tstr': {'type': 'string'},     # Text string (UTF-8)
    'text': {'type': 'string'},     # Alias for tstr

    # Boolean and null
    'bool': {'type': 'boolean'},
    'true': {'type': 'boolean'},
    'false': {'type': 'boolean'},
    'nil': {'type': 'null'},
    'null': {'type': 'null'},

    # Any type
    'any': {'type': 'any'},

    # Additional CBOR types
    'undefined': {'type': 'null'},  # CBOR undefined maps to null in JSON Structure
}


class CddlToStructureConverter:
    """
    Converts CDDL schema documents to JSON Structure format.
    """

    def __init__(self) -> None:
        self.root_namespace = DEFAULT_NAMESPACE
        self.type_registry: Dict[str, Dict[str, Any]] = {}
        self.definitions: Dict[str, Dict[str, Any]] = {}
        self.generic_types: Dict[str, Dict[str, Any]] = {}  # type name -> generic info
        self.current_generic_bindings: Dict[str, Dict[str, Any]] = {}  # type param bindings
        self.generic_template_names: Set[str] = set()  # generic templates (not concrete)
        # Cycle detection state
        self._conversion_stack: List[str] = []  # Stack of type names being converted
        self._conversion_depth: int = 0  # Current recursion depth
        # Counter for auto-generated type names
        self._auto_name_counter: int = 0

    def _create_inline_definition(
        self, type_def: Dict[str, Any], base_name: str
    ) -> Dict[str, Any]:
        """
        Move an inline compound type to definitions and return a $ref.

        JSON Structure requires compound types (object, tuple, array, map, etc.)
        inside unions to be referenced via $ref. This method handles the extraction.

        Args:
            type_def: The inline type definition to extract
            base_name: Base name to use for generating the definition name

        Returns:
            A $ref pointing to the new definition, or the original type if not compound
        """
        type_value = type_def.get('type')
        # All compound types that need extraction in union contexts
        if type_value not in ('object', 'tuple', 'choice', 'array', 'map', 'set'):
            return type_def

        # Generate a unique name for the inline type
        self._auto_name_counter += 1
        type_name = f"{base_name}_{type_value}_{self._auto_name_counter}"
        type_name = avro_name(type_name)

        # Add name to the type and store in definitions
        type_def['name'] = type_name
        self.definitions[type_name] = type_def

        # Return a $ref
        return {'$ref': f'#/definitions/{type_name}'}

    def _has_unresolved_refs(self, type_def: Dict[str, Any]) -> bool:
        """
        Check if a type definition has unresolved $ref pointers.

        Unresolved refs are typically generic type parameters (like T, E) that
        weren't substituted during instantiation.
        """
        def check(obj: Any) -> bool:
            if isinstance(obj, dict):
                if '$ref' in obj:
                    ref = obj['$ref']
                    # Check if this is a reference to a definitions entry
                    if ref.startswith('#/definitions/'):
                        ref_name = ref[14:]  # Remove '#/definitions/'
                        # Check if ref exists in definitions or type_registry
                        # Type registry has original names, definitions has transformed names
                        if ref_name in self.definitions:
                            return False  # Resolved
                        if ref_name in self.type_registry:
                            return False  # Resolved (original name)
                        # Check if the transformed name exists
                        if avro_name(ref_name) in self.definitions:
                            return False  # Resolved with transformed name
                        # Check for generic type parameters (typically single uppercase letters)
                        if len(ref_name) <= 2 and ref_name.isupper():
                            return True  # Likely an unresolved generic parameter
                        # If not found anywhere, it's unresolved
                        return True
                    elif ref.startswith('#/'):
                        # Other internal refs - allow them
                        return False
                for v in obj.values():
                    if check(v):
                        return True
            elif isinstance(obj, list):
                for item in obj:
                    if check(item):
                        return True
            return False
        return check(type_def)

    def _extract_description(self, node: Any) -> Optional[str]:
        """
        Extract description from comments attached to a node.

        CDDL comments start with ';' and are attached to AST nodes by the parser.
        This method extracts and cleans up those comments to use as descriptions.

        Args:
            node: Any AST node that might have comments attached

        Returns:
            A cleaned description string, or None if no comments
        """
        if not hasattr(node, 'comments') or not node.comments:
            return None

        # Extract comment text from each comment token
        comment_lines = []
        for comment in node.comments:
            if hasattr(comment, 'literal'):
                # Remove the leading semicolon and whitespace
                text = comment.literal.strip()
                if text.startswith(';'):
                    text = text[1:].strip()
                if text:
                    comment_lines.append(text)

        if not comment_lines:
            return None

        # Join multiple comment lines with spaces
        return ' '.join(comment_lines)

    def convert_cddl_to_structure(
        self, cddl_content: str, namespace: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Convert CDDL content to JSON Structure format.

        Args:
            cddl_content: The CDDL schema as a string
            namespace: Optional namespace for the schema

        Returns:
            Dict containing the JSON Structure schema
        """
        if namespace:
            self.root_namespace = namespace

        # Parse the CDDL content
        ast = cddl_parse(cddl_content)

        # Clear type registry and cycle detection state
        self.type_registry.clear()
        self.definitions.clear()
        self.generic_types.clear()
        self.current_generic_bindings.clear()
        self.generic_template_names.clear()
        self._conversion_stack.clear()
        self._conversion_depth = 0
        self._auto_name_counter = 0

        # First pass: collect all type names and generic definitions
        for rule in ast.rules:
            if hasattr(rule, 'getChildren'):
                children = rule.getChildren()
                if len(children) >= 2:
                    typename_node = children[0]
                    if hasattr(typename_node, 'name'):
                        self.type_registry[typename_node.name] = {}
                        # Check for generic parameters
                        generic_params = self._extract_generic_parameters(typename_node)
                        if generic_params:
                            self.generic_types[typename_node.name] = {
                                'params': generic_params,
                                'type_node': children[1] if len(children) > 1 else None
                            }
                            # Mark this as a generic template (not a concrete type)
                            self.generic_template_names.add(avro_name(typename_node.name))

        # Second pass: process all rules
        for rule in ast.rules:
            self._process_rule(rule)

        # Build the output structure
        structure_schema: Dict[str, Any] = {
            "$schema": "https://json-structure.org/meta/extended/v0/#",
            "$id": f"https://{self.root_namespace.replace('.', '/')}/schema.json"
        }

        # Add definitions if any (excluding generic templates and types with unresolved refs)
        concrete_definitions = {
            name: defn for name, defn in self.definitions.items()
            if name not in self.generic_template_names and not self._has_unresolved_refs(defn)
        }
        if concrete_definitions:
            structure_schema['definitions'] = concrete_definitions

        # If there's a root type (first rule that's not a generic template), add it to the schema
        if ast.rules and concrete_definitions:
            first_rule_name = self._get_rule_name(ast.rules[0])
            # Skip generic templates to find first concrete type
            for rule in ast.rules:
                first_rule_name = self._get_rule_name(rule)
                if first_rule_name and first_rule_name not in self.generic_template_names:
                    break
            if first_rule_name and first_rule_name in concrete_definitions:
                root_def = concrete_definitions[first_rule_name]
                # Copy root type properties to schema root
                for key, value in root_def.items():
                    if key not in structure_schema:
                        structure_schema[key] = value

        # Scan for extension usage
        uses = self._scan_for_uses(structure_schema)
        if uses:
            structure_schema['$uses'] = uses

        return structure_schema

    def _get_rule_name(self, rule: Rule) -> Optional[str]:
        """Extract the name from a rule."""
        if hasattr(rule, 'getChildren'):
            children = rule.getChildren()
            if len(children) >= 1:
                typename_node = children[0]
                if hasattr(typename_node, 'name'):
                    return avro_name(typename_node.name)
        return None

    def validate_structure_schema(self, structure_schema: Dict[str, Any],
                                   source_text: Optional[str] = None) -> List[ValidationError]:
        """
        Validate a JSON Structure schema using the json-structure SDK.

        Args:
            structure_schema: The JSON Structure schema to validate
            source_text: Optional source text for better error locations

        Returns:
            List of validation errors (empty if valid)

        Raises:
            CddlConversionError: If the schema is invalid
        """
        validator = SchemaValidator()
        errors = validator.validate(structure_schema, source_text)

        # Log any errors
        for error in errors:
            if error.severity == ValidationSeverity.ERROR:
                logger.error("Schema validation error: %s at %s", error.message, error.path)
            else:
                logger.warning("Schema validation warning: %s at %s", error.message, error.path)

        return errors

    def _process_rule(self, rule: Rule) -> None:
        """Process a CDDL rule and add it to definitions."""
        if not hasattr(rule, 'getChildren'):
            return

        children = rule.getChildren()
        if len(children) < 2:
            return

        # First child is the typename, second is the type
        typename_node = children[0]
        type_node = children[1] if len(children) > 1 else None

        if not hasattr(typename_node, 'name') or type_node is None:
            return

        rule_name = avro_name(typename_node.name)
        original_name = typename_node.name

        # Extract description from comments on the typename
        description = self._extract_description(typename_node)

        # Convert the type
        structure_type = self._convert_type(type_node, rule_name)

        if structure_type:
            if isinstance(structure_type, dict):
                structure_type['name'] = rule_name
                # Add description if present
                if description:
                    structure_type['description'] = description
                # Add altnames if original name differs
                if original_name != rule_name:
                    structure_type['altnames'] = {'cddl': original_name}
            self.definitions[rule_name] = structure_type
            self.type_registry[typename_node.name] = structure_type

    def _convert_type(self, type_node: Any, context_name: str = '') -> Dict[str, Any]:
        """
        Convert a CDDL type node to JSON Structure type.

        Args:
            type_node: The AST node to convert
            context_name: Name context for error messages and nested type naming

        Returns:
            JSON Structure type definition

        Raises:
            CddlConversionError: If conversion depth exceeds MAX_CONVERSION_DEPTH
        """
        if type_node is None:
            return {'type': 'any'}

        # Check recursion depth
        self._conversion_depth += 1
        if self._conversion_depth > MAX_CONVERSION_DEPTH:
            self._conversion_depth -= 1
            logger.warning("Maximum conversion depth exceeded for context: %s", context_name)
            raise CddlConversionError(
                f"Maximum conversion depth ({MAX_CONVERSION_DEPTH}) exceeded",
                context=context_name
            )

        try:
            node_type = type(type_node).__name__

            if node_type == 'Type':
                return self._convert_type_node(type_node, context_name)
            elif node_type == 'Typename':
                return self._convert_typename(type_node)
            elif node_type == 'Map':
                return self._convert_map(type_node, context_name)
            elif node_type == 'Array':
                return self._convert_array(type_node, context_name)
            elif node_type == 'Group':
                return self._convert_group(type_node, context_name)
            elif node_type == 'GroupChoice':
                return self._convert_group_choice(type_node, context_name)
            elif node_type == 'GroupEntry':
                return self._convert_group_entry(type_node, context_name)
            elif node_type == 'Value':
                return self._convert_value(type_node)
            elif node_type == 'Range':
                return self._convert_range(type_node)
            elif node_type == 'Operator':
                return self._convert_operator(type_node, context_name)
            elif node_type == 'Tag':
                return self._convert_tag(type_node, context_name)
            elif node_type == 'ChoiceFrom':
                return self._convert_choice_from(type_node, context_name)
            else:
                # Default fallback
                logger.debug("Unknown node type '%s', using 'any'", node_type)
                return {'type': 'any'}
        finally:
            self._conversion_depth -= 1

    def _convert_type_node(self, type_node: Type, context_name: str = '') -> Dict[str, Any]:
        """Convert a Type node."""
        if not hasattr(type_node, 'getChildren'):
            return {'type': 'any'}

        children = type_node.getChildren()
        if not children:
            return {'type': 'any'}

        # If there are multiple children, this might be a choice (union)
        if len(children) > 1:
            # Check if this is a choice type (type1 / type2)
            converted_types = []
            for child in children:
                converted = self._convert_type(child, context_name)
                if converted:
                    converted_types.append(converted)

            if len(converted_types) > 1:
                # Check if all types are string constants - if so, convert to enum
                all_string_consts = all(
                    isinstance(ct, dict) and
                    ct.get('type') == 'string' and
                    'const' in ct and
                    len(ct) == 2
                    for ct in converted_types
                )
                if all_string_consts:
                    # Convert to string enum
                    enum_values = [ct['const'] for ct in converted_types]
                    return {'type': 'string', 'enum': enum_values}

                # Create a union type per JSON Structure spec Section 3.5.1
                # Union elements should be type names (strings) or $ref schemas
                # Inline compound types must be extracted to definitions
                union_elements: List[Any] = []
                for ct in converted_types:
                    if '$ref' in ct:
                        # Keep $ref as-is (it's a valid union element)
                        union_elements.append(ct)
                    elif 'type' in ct and isinstance(ct['type'], str):
                        type_val = ct['type']
                        # Check if it's a compound type that needs extraction
                        # object, tuple, choice, array, map are all compound types
                        if type_val in ('object', 'tuple', 'choice', 'array', 'map', 'set'):
                            # Extract to definitions and use $ref
                            extracted = self._create_inline_definition(ct, f"{context_name}_option")
                            union_elements.append(extracted)
                        else:
                            # Check if it's a simple primitive type with no other attributes
                            other_keys = set(ct.keys()) - {'type'}
                            if not other_keys:
                                # Simple type - use just the type name
                                union_elements.append(type_val)
                            else:
                                # Type with constraints - keep full schema
                                union_elements.append(ct)
                    else:
                        # Keep as-is for other types
                        union_elements.append(ct)
                return {'type': union_elements}
            elif len(converted_types) == 1:
                return converted_types[0]

        # Single child
        return self._convert_type(children[0], context_name)

    def _convert_typename(self, typename_node: Typename) -> Dict[str, Any]:
        """Convert a Typename node."""
        if not hasattr(typename_node, 'name'):
            return {'type': 'any'}

        type_name = typename_node.name

        # Check for unwrap operator (~)
        has_unwrap = self._has_unwrap_operator(typename_node)

        # Check if this is a type parameter that's currently bound
        if type_name in self.current_generic_bindings:
            result = dict(self.current_generic_bindings[type_name])
            if has_unwrap:
                result = self._apply_unwrap(result)
            return result

        # Check if it's a primitive type
        if type_name in CDDL_PRIMITIVE_TYPES:
            return dict(CDDL_PRIMITIVE_TYPES[type_name])

        # Check for generic arguments
        generic_args = self._extract_generic_arguments(typename_node)

        # Check if this is a reference to a generic type with arguments
        if generic_args and type_name in self.generic_types:
            result = self._instantiate_generic_type(type_name, generic_args)
            if has_unwrap:
                result = self._apply_unwrap(result)
            return result

        # Check if it's a reference to another type
        normalized_name = avro_name(type_name)
        ref_result: Dict[str, Any]
        if type_name in self.type_registry or normalized_name in self.definitions:
            ref_result = {'$ref': f'#/definitions/{normalized_name}'}
        else:
            # Unknown type - treat as reference
            ref_result = {'$ref': f'#/definitions/{normalized_name}'}

        if has_unwrap:
            ref_result = self._apply_unwrap(ref_result)
        return ref_result

    def _has_unwrap_operator(self, typename_node: Any) -> bool:
        """Check if a typename node has the unwrap operator (~)."""
        # Check for 'unwrapped' attribute which contains the TILDE token
        if hasattr(typename_node, 'unwrapped') and typename_node.unwrapped is not None:
            return True
        # Fallback: check children for TILDE token
        if hasattr(typename_node, 'getChildren'):
            for child in typename_node.getChildren():
                if hasattr(child, 'kind') and 'TILDE' in str(child.kind):
                    return True
        return False

    def _apply_unwrap(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """
        Apply the unwrap operator (~) to a type.

        The unwrap operator extracts the content from a group or array.
        In JSON Structure, we mark this with a special annotation.
        """
        # For referenced types, we need to resolve and unwrap
        if '$ref' in type_def:
            ref_name = type_def['$ref'].split('/')[-1]
            if ref_name in self.definitions:
                resolved = self.definitions[ref_name]
                # If it's an object, return its properties inline
                if resolved.get('type') == 'object' and 'properties' in resolved:
                    return dict(resolved)
                # If it's an array, return items type
                if resolved.get('type') == 'array' and 'items' in resolved:
                    return resolved['items']

        # For inline types, just return as-is (unwrap is contextual)
        return type_def

    def _extract_generic_parameters(self, typename_node: Any) -> List[str]:
        """Extract generic type parameters from a typename node (e.g., optional<T> -> ['T'])."""
        params = []
        if hasattr(typename_node, 'getChildren'):
            for child in typename_node.getChildren():
                if type(child).__name__ == 'GenericParameters':
                    for pc in child.getChildren():
                        if type(pc).__name__ == 'Typename' and hasattr(pc, 'name'):
                            params.append(pc.name)
        return params

    def _extract_generic_arguments(self, typename_node: Any) -> List[Any]:
        """Extract generic type arguments from a typename node.

        Returns AST nodes for the arguments so they can be properly converted
        with current bindings in _convert_typename.
        """
        args = []
        if hasattr(typename_node, 'getChildren'):
            for child in typename_node.getChildren():
                if type(child).__name__ == 'GenericArguments':
                    for ac in child.getChildren():
                        # Return the AST node itself for proper conversion
                        args.append(ac)
        return args

    def _resolve_generic_argument(self, arg_node: Any) -> Dict[str, Any]:
        """Resolve a generic argument node to a JSON Structure type."""
        if type(arg_node).__name__ == 'Typename':
            if hasattr(arg_node, 'name'):
                arg_name = arg_node.name
                # Check if it's a bound type parameter
                if arg_name in self.current_generic_bindings:
                    return dict(self.current_generic_bindings[arg_name])
                # Check for primitives
                if arg_name in CDDL_PRIMITIVE_TYPES:
                    return dict(CDDL_PRIMITIVE_TYPES[arg_name])
                # Otherwise it's a reference to a defined type
                normalized_name = avro_name(arg_name)
                return {'$ref': f'#/definitions/{normalized_name}'}
        # Fallback: try to convert it
        return self._convert_type(arg_node, '')

    def _instantiate_generic_type(
        self, type_name: str, type_arg_nodes: List[Any]
    ) -> Dict[str, Any]:
        """
        Instantiate a generic type with concrete type arguments.

        Args:
            type_name: Name of the generic type (e.g., 'optional', 'pair')
            type_arg_nodes: AST nodes for the type arguments

        Returns:
            JSON Structure type definition with type arguments substituted

        Raises:
            CddlCycleError: If a circular generic instantiation is detected
        """
        generic_info = self.generic_types.get(type_name)
        if not generic_info:
            return {'type': 'any'}

        # Check for cycles in generic instantiation
        if type_name in self._conversion_stack:
            cycle_start = self._conversion_stack.index(type_name)
            cycle_path = self._conversion_stack[cycle_start:] + [type_name]
            logger.warning(
                "Circular generic instantiation detected: %s", ' -> '.join(cycle_path)
            )
            # Return a reference instead of raising to allow recursive types
            normalized_name = avro_name(type_name)
            return {'$ref': f'#/definitions/{normalized_name}'}

        params = generic_info.get('params', [])
        type_node = generic_info.get('type_node')

        if not type_node or len(params) != len(type_arg_nodes):
            # Parameter count mismatch - return reference
            normalized_name = avro_name(type_name)
            return {'$ref': f'#/definitions/{normalized_name}'}

        # Resolve type arguments with current bindings before creating new bindings
        resolved_args = [self._resolve_generic_argument(arg) for arg in type_arg_nodes]

        # Create bindings for type parameters
        old_bindings = self.current_generic_bindings.copy()
        for param, arg in zip(params, resolved_args):
            self.current_generic_bindings[param] = arg

        # Push to conversion stack for cycle detection
        self._conversion_stack.append(type_name)
        try:
            # Convert the type with the current bindings
            result = self._convert_type(type_node, type_name)
            return result
        finally:
            # Pop from conversion stack and restore previous bindings
            self._conversion_stack.pop()
            self.current_generic_bindings = old_bindings

    def _convert_map(self, map_node: Map, context_name: str = '') -> Dict[str, Any]:
        """Convert a Map node to JSON Structure object or map."""
        properties: Dict[str, Any] = {}
        required: List[str] = []
        extends_refs: List[str] = []
        computed_key_info: Optional[Dict[str, Any]] = None

        # Process map contents
        if hasattr(map_node, 'getChildren'):
            for child in map_node.getChildren():
                child_type = type(child).__name__
                if child_type == 'GroupChoice':
                    computed_key_info = self._process_group_choice_for_object(
                        child, properties, required, extends_refs, context_name
                    )

        # If we have a computed key (like * tstr => int) and no explicit properties,
        # this is a JSON Structure map type
        if computed_key_info and not properties and not extends_refs:
            result: Dict[str, Any] = {'type': 'map'}
            if computed_key_info.get('keys'):
                result['keys'] = computed_key_info['keys']
            if computed_key_info.get('values'):
                result['values'] = computed_key_info['values']
            return result

        # Otherwise it's a regular object
        result = {'type': 'object'}
        if extends_refs:
            # Use $extends for unwrapped types - only first one supported
            # JSON Structure only supports single inheritance via $extends
            result['$extends'] = extends_refs[0]
            if len(extends_refs) > 1:
                logger.warning(
                    "Multiple unwrap operators found, only using first: %s. "
                    "JSON Structure $extends only supports single inheritance.",
                    extends_refs[0]
                )
        if properties:
            result['properties'] = properties
        if required:
            result['required'] = required

        return result

    def _process_group_choice_for_object(
        self,
        group_choice: CDDLNode,
        properties: Dict[str, Any],
        required: List[str],
        extends_refs: List[str],
        context_name: str
    ) -> Optional[Dict[str, Any]]:
        """Process GroupChoice for object properties.

        Returns computed key info if a computed key entry (like * tstr => int) is found.
        """
        if not hasattr(group_choice, 'getChildren'):
            return None

        computed_key_info: Optional[Dict[str, Any]] = None

        for child in group_choice.getChildren():
            child_type = type(child).__name__
            if child_type == 'GroupEntry':
                entry_computed = self._process_group_entry_for_object(
                    child, properties, required, extends_refs, context_name
                )
                if entry_computed:
                    computed_key_info = entry_computed
            elif child_type == 'GroupChoice':
                # Nested group choice
                nested_computed = self._process_group_choice_for_object(
                    child, properties, required, extends_refs, context_name
                )
                if nested_computed:
                    computed_key_info = nested_computed

        return computed_key_info

    def _process_group_entry_for_object(
        self,
        entry: CDDLNode,
        properties: Dict[str, Any],
        required: List[str],
        extends_refs: List[str],
        context_name: str
    ) -> Optional[Dict[str, Any]]:
        """Process a GroupEntry for object properties.

        Returns computed key info if this is a computed key entry (like * tstr => int).
        """
        if not hasattr(entry, 'getChildren'):
            return None

        children = entry.getChildren()

        # Parse occurrence, memberkey, and type from children
        occurrence_indicator = None
        member_key = None
        member_type = None
        unwrap_typename = None

        for child in children:
            child_type = type(child).__name__
            if child_type == 'Occurrence':
                occurrence_indicator = self._parse_occurrence(child)
            elif child_type == 'Memberkey':
                member_key = self._parse_memberkey(child)
            elif child_type == 'Typename':
                # Check if this is an unwrap operator usage (~typename)
                if self._has_unwrap_operator(child):
                    unwrap_typename = child
                else:
                    member_type = self._convert_type(child, context_name)
            elif child_type == 'Type':
                # Check if this Type contains an unwrapped Typename
                if hasattr(child, 'getChildren'):
                    type_children = child.getChildren()
                    if len(type_children) == 1 and type(type_children[0]).__name__ == 'Typename':
                        typename = type_children[0]
                        if self._has_unwrap_operator(typename):
                            unwrap_typename = typename
                            continue
                member_type = self._convert_type(child, context_name)
            elif child_type in ('Map', 'Array', 'Group', 'Value'):
                member_type = self._convert_type(child, context_name)

        # Handle unwrap operator (~) - use $extends to reference the base type
        if unwrap_typename is not None and member_key is None:
            ref_name = getattr(unwrap_typename, 'name', None)
            if ref_name:
                normalized_ref = avro_name(ref_name)
                # Add $ref to extends_refs for $extends
                ref = f"#/definitions/{normalized_ref}"
                if ref not in extends_refs:
                    extends_refs.append(ref)
            return None

        if member_key is None:
            return None

        prop_name = member_key.get('name', '')
        original_name = member_key.get('original_name', prop_name)
        is_computed = member_key.get('computed', False)

        # Handle computed keys (patterns like * tstr => int)
        if is_computed:
            key_type_name = member_key.get('key_type', 'string')
            # Map CDDL key type to JSON Structure type
            if key_type_name in ('tstr', 'text'):
                keys_type = {'type': 'string'}
            elif key_type_name in ('bstr', 'bytes'):
                keys_type = {'type': 'binary'}
            elif key_type_name in ('int', 'uint', 'nint'):
                keys_type = {'type': 'int64'}
            else:
                keys_type = {'type': 'string'}

            values_type = member_type if member_type else {'type': 'any'}

            return {
                'keys': keys_type,
                'values': values_type
            }

        if not prop_name:
            return None

        normalized_name = avro_name(prop_name)

        if member_type:
            prop_schema = member_type.copy() if isinstance(member_type, dict) else member_type
        else:
            prop_schema = {'type': 'any'}

        # Handle occurrence indicators
        if occurrence_indicator:
            if occurrence_indicator.get('optional'):
                # Optional field - no need to add to required
                pass
            elif occurrence_indicator.get('min', 1) >= 1:
                required.append(normalized_name)

            # Handle array occurrences (* or +)
            if occurrence_indicator.get('array'):
                prop_schema = {
                    'type': 'array',
                    'items': prop_schema
                }
                if occurrence_indicator.get('min'):
                    prop_schema['minItems'] = occurrence_indicator['min']
                if occurrence_indicator.get('max'):
                    prop_schema['maxItems'] = occurrence_indicator['max']
        else:
            # No occurrence indicator means required
            required.append(normalized_name)

        # Add altnames if original name differs
        if original_name != normalized_name:
            if isinstance(prop_schema, dict):
                prop_schema['altnames'] = {'cddl': original_name}

        # Add description if present in member_key or occurrence indicator
        if isinstance(prop_schema, dict):
            desc = member_key.get('description')
            if not desc and occurrence_indicator:
                desc = occurrence_indicator.get('description')
            if desc:
                prop_schema['description'] = desc

        properties[normalized_name] = prop_schema
        return None

    def _parse_occurrence(self, occurrence: CDDLNode) -> Dict[str, Any]:
        """Parse an Occurrence node."""
        result: Dict[str, Any] = {}

        if not hasattr(occurrence, 'getChildren'):
            return {'optional': True}

        # Check for comments on the occurrence tokens
        if hasattr(occurrence, 'tokens') and occurrence.tokens:
            for token in occurrence.tokens:
                if hasattr(token, 'comments') and token.comments:
                    description = self._extract_description_from_comments(token.comments)
                    if description:
                        result['description'] = description
                        break

        for child in occurrence.getChildren():
            # Check for occurrence tokens
            if hasattr(child, 'kind'):
                token_kind = str(child.kind)
                if 'QUEST' in token_kind:  # ?
                    result['optional'] = True
                elif 'ASTERISK' in token_kind:  # *
                    result['array'] = True
                    result['min'] = 0
                elif 'PLUS' in token_kind:  # +
                    result['array'] = True
                    result['min'] = 1
            elif hasattr(child, 'value'):
                # Handle numeric occurrences like n*m
                pass

        if not result or ('optional' not in result and 'array' not in result):
            result['optional'] = True

        return result

    def _extract_description_from_comments(self, comments: List[Any]) -> Optional[str]:
        """Extract description from a list of comment tokens."""
        if not comments:
            return None

        comment_lines = []
        for comment in comments:
            if hasattr(comment, 'literal'):
                text = comment.literal.strip()
                if text.startswith(';'):
                    text = text[1:].strip()
                if text:
                    comment_lines.append(text)

        return ' '.join(comment_lines) if comment_lines else None

    def _parse_memberkey(self, memberkey: CDDLNode) -> Optional[Dict[str, Any]]:
        """Parse a Memberkey node."""
        if not hasattr(memberkey, 'getChildren'):
            return None

        result: Dict[str, Any] = {}

        for child in memberkey.getChildren():
            child_type = type(child).__name__
            if child_type == 'Typename':
                if hasattr(child, 'name'):
                    name = child.name
                    # Check if this is a computed key (like tstr => any)
                    if name in ('tstr', 'text', 'bstr', 'bytes', 'int', 'uint'):
                        result['computed'] = True
                        result['key_type'] = name
                    else:
                        result['name'] = avro_name(name)
                        result['original_name'] = name
                        # Extract description from comments on the typename
                        description = self._extract_description(child)
                        if description:
                            result['description'] = description
            elif child_type == 'Value':
                # Could be string key or integer key
                if hasattr(child, 'value'):
                    value = child.value
                    # Remove quotes if present (string key)
                    if value.startswith('"') and value.endswith('"'):
                        value = value[1:-1]
                        result['name'] = avro_name(value)
                        result['original_name'] = value
                    else:
                        # Try to parse as integer key
                        try:
                            int_key = int(value)
                            result['name'] = f'_{value}'
                            result['original_name'] = value
                            result['integer_key'] = int_key
                        except ValueError:
                            result['name'] = avro_name(value)
                            result['original_name'] = value

        return result if result else None

    def _convert_array(self, array_node: Array, context_name: str = '') -> Dict[str, Any]:
        """Convert an Array node to JSON Structure array or tuple."""
        items_types: List[Dict[str, Any]] = []
        is_tuple = False

        if hasattr(array_node, 'getChildren'):
            for child in array_node.getChildren():
                child_type = type(child).__name__
                if child_type == 'GroupChoice':
                    # Get the items types from the group
                    items_types, is_tuple = self._get_array_items_types(child, context_name)
                elif child_type in ('Type', 'Typename'):
                    items_types = [self._convert_type(child, context_name)]

        if is_tuple and len(items_types) > 1:
            # Fixed-length array with different types = tuple
            # Use the JSON Structure tuple format
            properties: Dict[str, Any] = {}
            tuple_order: List[str] = []
            for idx, item_type in enumerate(items_types):
                prop_name = f"_{idx}"
                properties[prop_name] = item_type
                tuple_order.append(prop_name)
            return {
                'type': 'tuple',
                'properties': properties,
                'tuple': tuple_order
            }
        elif items_types:
            # Regular array
            if len(items_types) == 1:
                items_type = items_types[0]
                # Extract inline compound types to definitions
                items_type = self._create_inline_definition(items_type, f"{context_name}_item")
            else:
                # Multiple types that aren't a tuple - use union for items
                # Extract each inline compound type to definitions
                extracted_types = [
                    self._create_inline_definition(t, f"{context_name}_item")
                    for t in items_types
                ]
                items_type = {'type': extracted_types}
            return {
                'type': 'array',
                'items': items_type
            }
        else:
            return {
                'type': 'array',
                'items': {'type': 'any'}
            }

    def _get_array_items_types(
        self, group_choice: CDDLNode, context_name: str
    ) -> Tuple[List[Dict[str, Any]], bool]:
        """Extract items types from a GroupChoice in an array context.

        Returns:
            Tuple of (list of item types, is_tuple flag)
            is_tuple is True if this looks like a fixed-size tuple definition
        """
        if not hasattr(group_choice, 'getChildren'):
            return [{'type': 'any'}], False

        children = group_choice.getChildren()
        if not children:
            return [{'type': 'any'}], False

        # Collect all types from group entries
        types = []
        all_entries_are_single = True

        for child in children:
            child_type = type(child).__name__
            if child_type == 'GroupEntry':
                entry_type = self._get_group_entry_type(child, context_name)
                if entry_type:
                    types.append(entry_type)
                # Check if entry has occurrence markers (*, +, ?)
                if self._has_occurrence_marker(child):
                    all_entries_are_single = False
            elif child_type in ('Type', 'Typename'):
                types.append(self._convert_type(child, context_name))

        # It's a tuple if we have multiple distinct types AND no occurrence markers
        is_tuple = len(types) > 1 and all_entries_are_single

        return types, is_tuple

    def _has_occurrence_marker(self, entry: CDDLNode) -> bool:
        """Check if a GroupEntry has occurrence markers (*, +, ?)."""
        if hasattr(entry, 'getChildren'):
            for child in entry.getChildren():
                if type(child).__name__ == 'Occurrence':
                    return True
        return False

    def _get_group_entry_type(self, entry: CDDLNode, context_name: str) -> Optional[Dict[str, Any]]:
        """Get the type from a GroupEntry."""
        if not hasattr(entry, 'getChildren'):
            return None

        for child in entry.getChildren():
            child_type = type(child).__name__
            if child_type in ('Type', 'Typename', 'Map', 'Array', 'Group', 'Value'):
                return self._convert_type(child, context_name)

        return None

    def _convert_group(self, group_node: CDDLNode, context_name: str = '') -> Dict[str, Any]:
        """Convert a Group node."""
        # Check if this is an integer-keyed group (should become a tuple)
        int_key_entries = self._extract_integer_key_entries(group_node)
        if int_key_entries:
            return self._convert_to_tuple(int_key_entries, context_name)

        # Check if this is a sequence of unnamed entries (should become a tuple)
        unnamed_entries = self._extract_unnamed_entries(group_node)
        if unnamed_entries:
            # Convert to tuple format - assign sequential indices
            indexed_entries = [(i, False, entry) for i, entry in enumerate(unnamed_entries)]
            return self._convert_to_tuple(indexed_entries, context_name)

        # Otherwise, treat as an object
        return self._convert_map_like(group_node, context_name)

    def _extract_unnamed_entries(self, node: Any) -> Optional[List[Any]]:
        """Extract unnamed entries from a group (entries with only types, no member keys)."""
        entries: List[Any] = []
        has_named = False

        if not hasattr(node, 'getChildren'):
            return None

        for child in node.getChildren():
            if type(child).__name__ == 'GroupChoice':
                for gc_child in child.getChildren():
                    if type(gc_child).__name__ == 'GroupEntry':
                        has_memberkey = False
                        type_node = None
                        for entry_child in gc_child.getChildren():
                            child_type = type(entry_child).__name__
                            if child_type == 'Memberkey':
                                has_memberkey = True
                            elif child_type in ('Type', 'Typename'):
                                type_node = entry_child
                        if has_memberkey:
                            has_named = True
                        elif type_node:
                            entries.append(type_node)

        # Only return if all entries are unnamed
        if entries and not has_named:
            return entries
        return None

    def _extract_integer_key_entries(self, node: Any) -> Optional[List[Tuple[int, bool, Any]]]:
        """
        Extract integer-keyed entries from a group.
        Returns list of (key, is_optional, type_node) tuples, or None if not integer-keyed.
        """
        entries: List[Tuple[int, bool, Any]] = []

        if not hasattr(node, 'getChildren'):
            return None

        for child in node.getChildren():
            if type(child).__name__ == 'GroupChoice':
                result = self._extract_int_keys_from_group_choice(child)
                if result is None:
                    return None  # Mixed keys, not a pure integer-keyed group
                entries.extend(result)

        if not entries:
            return None

        # Sort by key and check for sequential keys starting from 1
        entries.sort(key=lambda x: x[0])

        return entries

    def _extract_int_keys_from_group_choice(
        self, group_choice: Any
    ) -> Optional[List[Tuple[int, bool, Any]]]:
        """Extract integer keys from a GroupChoice node."""
        entries: List[Tuple[int, bool, Any]] = []

        if not hasattr(group_choice, 'getChildren'):
            return None

        for child in group_choice.getChildren():
            if type(child).__name__ == 'GroupEntry':
                result = self._extract_int_key_from_entry(child)
                if result is None:
                    return None  # Not an integer key
                entries.append(result)
            elif type(child).__name__ == 'GroupChoice':
                nested = self._extract_int_keys_from_group_choice(child)
                if nested is None:
                    return None
                entries.extend(nested)

        return entries

    def _extract_int_key_from_entry(self, entry: Any) -> Optional[Tuple[int, bool, Any]]:
        """Extract integer key, optionality, and type from a GroupEntry."""
        if not hasattr(entry, 'getChildren'):
            return None

        children = entry.getChildren()
        is_optional = False
        int_key = None
        type_node = None

        for child in children:
            child_type = type(child).__name__
            if child_type == 'Occurrence':
                # Check for optional marker
                is_optional = self._is_optional_occurrence(child)
            elif child_type == 'Memberkey':
                int_key = self._get_integer_key(child)
                if int_key is None:
                    return None  # Not an integer key
            elif child_type in ('Type', 'Typename', 'Map', 'Array', 'Group', 'Value'):
                type_node = child

        if int_key is not None and type_node is not None:
            return (int_key, is_optional, type_node)

        return None

    def _is_optional_occurrence(self, occurrence: Any) -> bool:
        """Check if an occurrence node indicates optional (?)."""
        # The ? marker means n=0, m=1 (zero or one)
        # Check the n attribute - if n=0, it's optional
        if hasattr(occurrence, 'n') and occurrence.n == 0:
            return True
        # Fallback: check tokens for QUEST
        if hasattr(occurrence, 'tokens'):
            for token in occurrence.tokens:
                if 'QUEST' in repr(token):
                    return True
        return False

    def _get_integer_key(self, memberkey: Any) -> Optional[int]:
        """Get integer key from a Memberkey node, or None if not an integer key."""
        if not hasattr(memberkey, 'getChildren'):
            return None

        for child in memberkey.getChildren():
            if type(child).__name__ == 'Value' and hasattr(child, 'value'):
                try:
                    return int(child.value)
                except ValueError:
                    return None

        return None

    def _convert_to_tuple(
        self, entries: List[Tuple[int, bool, Any]], context_name: str
    ) -> Dict[str, Any]:
        """Convert integer-keyed entries to a tuple type per JSON Structure spec.

        JSON Structure tuples use:
        - 'properties': named properties with schemas
        - 'tuple': array of property names defining the order
        - All declared properties are implicitly REQUIRED

        For optional elements, we omit them from the 'tuple' array but keep them
        in properties, or we can include them and note that JSON Structure tuples
        don't support optional elements directly.
        """
        # Sort entries by key
        entries.sort(key=lambda x: x[0])

        # Build properties map and tuple order array
        properties: Dict[str, Any] = {}
        tuple_order: List[str] = []

        for key, is_optional, type_node in entries:
            # Generate property name from the integer key
            prop_name = f"_{key}"
            item_type = self._convert_type(type_node, context_name)

            # Store the original CDDL key as altname
            if isinstance(item_type, dict):
                item_type['altnames'] = {'cddl': str(key)}

            properties[prop_name] = item_type

            # Only include non-optional items in tuple order
            # (optional items can still be present but are not required)
            if not is_optional:
                tuple_order.append(prop_name)

        result: Dict[str, Any] = {
            'type': 'tuple',
            'properties': properties,
            'tuple': [f"_{e[0]}" for e in entries]  # Include all in order for the tuple layout
        }

        # If some items are optional, add required to indicate which are mandatory
        if tuple_order and len(tuple_order) < len(entries):
            result['required'] = tuple_order

        return result

    def _convert_map_like(self, node: Any, context_name: str = '') -> Dict[str, Any]:
        """Convert a map-like structure to object."""
        properties: Dict[str, Any] = {}
        required: List[str] = []
        extends_refs: List[str] = []

        if hasattr(node, 'getChildren'):
            for child in node.getChildren():
                child_type = type(child).__name__
                if child_type == 'GroupChoice':
                    self._process_group_choice_for_object(
                        child, properties, required, extends_refs, context_name
                    )

        result: Dict[str, Any] = {'type': 'object'}
        if extends_refs:
            # JSON Structure only supports single inheritance via $extends
            result['$extends'] = extends_refs[0]
            if len(extends_refs) > 1:
                logger.warning(
                    "Multiple unwrap operators found, only using first: %s. "
                    "JSON Structure $extends only supports single inheritance.",
                    extends_refs[0]
                )
        if properties:
            result['properties'] = properties
        if required:
            result['required'] = required

        return result

    def _convert_group_entry(self, entry: CDDLNode, context_name: str = '') -> Dict[str, Any]:
        """Convert a GroupEntry node used as a top-level type (e.g., for inline groups)."""
        # A GroupEntry at top level typically contains a Type with a Group inside
        if not hasattr(entry, 'getChildren'):
            return {'type': 'any'}

        for child in entry.getChildren():
            child_type = type(child).__name__
            if child_type == 'Type':
                return self._convert_type(child, context_name)
            elif child_type == 'Group':
                return self._convert_group(child, context_name)
            elif child_type == 'GroupChoice':
                return self._convert_group_choice(child, context_name)

        return {'type': 'any'}

    def _convert_group_choice(
        self, group_choice: CDDLNode, context_name: str = ''
    ) -> Dict[str, Any]:
        """Convert a GroupChoice node."""
        # Process as an object with properties
        properties: Dict[str, Any] = {}
        required: List[str] = []
        extends_refs: List[str] = []

        self._process_group_choice_for_object(
            group_choice, properties, required, extends_refs, context_name
        )

        result: Dict[str, Any] = {'type': 'object'}
        if extends_refs:
            # JSON Structure only supports single inheritance via $extends
            result['$extends'] = extends_refs[0]
            if len(extends_refs) > 1:
                logger.warning(
                    "Multiple unwrap operators found, only using first: %s. "
                    "JSON Structure $extends only supports single inheritance.",
                    extends_refs[0]
                )
        if properties:
            result['properties'] = properties
        if required:
            result['required'] = required

        return result

    def _convert_value(self, value_node: Value) -> Dict[str, Any]:
        """Convert a Value node (literal value)."""
        if not hasattr(value_node, 'value'):
            return {'type': 'any'}

        value = value_node.value

        # Determine type based on value
        if value.startswith('"') and value.endswith('"'):
            # String literal - create enum with single value
            string_val = value[1:-1]
            return {
                'type': 'string',
                'const': string_val
            }
        elif value in ('true', 'false'):
            return {
                'type': 'boolean',
                'const': value == 'true'
            }
        elif value.isdigit() or (value.startswith('-') and value[1:].isdigit()):
            return {
                'type': 'int64',
                'const': int(value)
            }
        elif '.' in value:
            try:
                float_val = float(value)
                return {
                    'type': 'double',
                    'const': float_val
                }
            except ValueError:
                pass

        return {'type': 'string', 'const': value}

    def _convert_range(self, range_node: Range) -> Dict[str, Any]:
        """Convert a Range node (min..max or min...max)."""
        result: Dict[str, Any] = {'type': 'int64'}

        if hasattr(range_node, 'getChildren'):
            children = range_node.getChildren()
            values: List[Any] = []
            for child in children:
                if hasattr(child, 'value'):
                    try:
                        values.append(int(child.value))
                    except ValueError:
                        try:
                            values.append(float(child.value))
                            result['type'] = 'double'
                        except ValueError:
                            pass

            if len(values) >= 2:
                result['minimum'] = values[0]
                result['maximum'] = values[1]
            elif len(values) == 1:
                result['minimum'] = values[0]

        return result

    def _convert_operator(self, operator_node: Operator, context_name: str = '') -> Dict[str, Any]:
        """
        Convert an Operator node (type constraints like .size, .regexp, .default).

        Supported control operators:
        - .size: Maps to minLength/maxLength for strings, minItems/maxItems for arrays
        - .regexp: Maps to pattern (ECMAScript regex)
        - .default: Maps to default value
        - .bits, .cbor, .cborseq: Base type extracted (CBOR-specific, no JSON Structure equivalent)
        - .within, .and: Type intersection (limited support)

        See RFC 8610 Section 3.8 for full control operator specification.
        """
        base_type: Dict[str, Any] = {'type': 'any'}
        operator_name: Optional[str] = None
        controller_value: Any = None

        # Extract operator name and base type
        if hasattr(operator_node, 'name') and hasattr(operator_node.name, 'literal'):
            operator_name = operator_node.name.literal

        if hasattr(operator_node, 'type'):
            # Use _convert_type which handles all node types properly
            if operator_node.type:
                base_type = self._convert_type(operator_node.type, context_name)
            else:
                base_type = {'type': 'any'}

        # Extract controller (constraint argument)
        if hasattr(operator_node, 'controller'):
            controller_value = self._extract_controller_value(operator_node.controller)

        # Apply operator-specific mappings
        if operator_name == 'size':
            base_type = self._apply_size_constraint(base_type, controller_value)
        elif operator_name == 'regexp':
            base_type = self._apply_regexp_constraint(base_type, controller_value)
        elif operator_name == 'default':
            base_type = self._apply_default_value(base_type, controller_value)
        elif operator_name == 'lt':
            # .lt (less than) -> exclusiveMaximum
            if isinstance(controller_value, (int, float)):
                base_type['exclusiveMaximum'] = controller_value
        elif operator_name == 'le':
            # .le (less or equal) -> maximum
            if isinstance(controller_value, (int, float)):
                base_type['maximum'] = controller_value
        elif operator_name == 'gt':
            # .gt (greater than) -> exclusiveMinimum
            if isinstance(controller_value, (int, float)):
                base_type['exclusiveMinimum'] = controller_value
        elif operator_name == 'ge':
            # .ge (greater or equal) -> minimum
            if isinstance(controller_value, (int, float)):
                base_type['minimum'] = controller_value
        elif operator_name == 'eq':
            # .eq (equals) -> const
            base_type['const'] = controller_value
        elif operator_name == 'ne':
            # .ne (not equals) - no direct JSON Structure equivalent, add as comment
            pass
        elif operator_name in ('within', 'and'):
            # Type intersection - limited support, just use base type
            pass
        elif operator_name in ('bits', 'cbor', 'cborseq'):
            # CBOR-specific operators - just use base type
            pass

        return base_type

    def _extract_controller_value(self, controller: Any) -> Any:
        """Extract the value from an operator's controller (argument)."""
        if controller is None:
            return None

        controller_type = type(controller).__name__

        # Handle direct Value node (e.g., .size 32, .ge 0)
        if controller_type == 'Value':
            return self._parse_value_literal(controller)

        # Handle direct Range node
        if controller_type == 'Range':
            return self._extract_range_values(controller)

        # Handle direct Typename node
        if controller_type == 'Typename' and hasattr(controller, 'name'):
            return controller.name

        # Handle Type node containing children (e.g., .size (1..100))
        if hasattr(controller, 'getChildren'):
            children = controller.getChildren()
            for child in children:
                child_type = type(child).__name__
                if child_type == 'Range':
                    # Extract min/max from range
                    return self._extract_range_values(child)
                if child_type == 'Value':
                    return self._parse_value_literal(child)
                if child_type == 'Typename':
                    if hasattr(child, 'name'):
                        return child.name

        return None

    def _extract_range_values(self, range_node: Any) -> Dict[str, Any]:
        """Extract min and max from a Range node."""
        result: Dict[str, Any] = {}
        values = []
        is_exclusive = False

        if hasattr(range_node, 'getChildren'):
            for child in range_node.getChildren():
                child_type = type(child).__name__
                if child_type == 'Value':
                    values.append(self._parse_value_literal(child))
                elif hasattr(child, 'kind') and 'EXCLRANGE' in str(child.kind):
                    is_exclusive = True

        if len(values) >= 2:
            result['min'] = values[0]
            result['max'] = values[1]
            result['exclusive'] = is_exclusive
        elif len(values) == 1:
            result['min'] = values[0]
            result['exclusive'] = is_exclusive

        return result

    def _parse_value_literal(self, value_node: Any) -> Any:
        """Parse a Value node and return the Python value."""
        if not hasattr(value_node, 'value'):
            return None

        value = value_node.value

        # String value
        if value.startswith('"') and value.endswith('"'):
            return value[1:-1]

        # Boolean
        if value == 'true':
            return True
        if value == 'false':
            return False

        # Integer
        try:
            return int(value)
        except ValueError:
            pass

        # Float
        try:
            return float(value)
        except ValueError:
            pass

        return value

    def _apply_size_constraint(self, base_type: Dict[str, Any], constraint: Any) -> Dict[str, Any]:
        """Apply .size constraint to a type."""
        if constraint is None:
            return base_type

        type_name = base_type.get('type', 'any')

        # Determine if this is a string or array type
        is_string_type = type_name in ('string', 'binary')
        is_array_type = type_name == 'array'

        if isinstance(constraint, dict):
            # Range constraint
            min_val = constraint.get('min')
            max_val = constraint.get('max')

            if is_string_type:
                if min_val is not None:
                    base_type['minLength'] = min_val
                if max_val is not None:
                    base_type['maxLength'] = max_val
            elif is_array_type:
                if min_val is not None:
                    base_type['minItems'] = min_val
                if max_val is not None:
                    base_type['maxItems'] = max_val
            else:
                # For other types, might be constraining bit/byte size
                # Map to general min/max
                if min_val is not None:
                    base_type['minimum'] = min_val
                if max_val is not None:
                    base_type['maximum'] = max_val
        elif isinstance(constraint, int):
            # Exact size constraint
            if is_string_type:
                base_type['minLength'] = constraint
                base_type['maxLength'] = constraint
            elif is_array_type:
                base_type['minItems'] = constraint
                base_type['maxItems'] = constraint

        return base_type

    def _apply_regexp_constraint(self, base_type: Dict[str, Any], pattern: Any) -> Dict[str, Any]:
        """Apply .regexp constraint to a string type."""
        if pattern and isinstance(pattern, str):
            base_type['pattern'] = pattern
        return base_type

    def _apply_default_value(self, base_type: Dict[str, Any], default_val: Any) -> Dict[str, Any]:
        """Apply .default value to a type."""
        if default_val is not None:
            base_type['default'] = default_val
        return base_type

    def _convert_tag(self, tag_node: Tag, context_name: str = '') -> Dict[str, Any]:
        """Convert a Tag node (CBOR tags like #6.n)."""
        # CBOR tags don't have a direct JSON Structure equivalent
        # Just convert the underlying type
        if hasattr(tag_node, 'getChildren'):
            for child in tag_node.getChildren():
                child_type = type(child).__name__
                if child_type in ('Type', 'Typename', 'Map', 'Array'):
                    return self._convert_type(child, context_name)

        return {'type': 'any'}

    def _convert_choice_from(
        self, choice_from: ChoiceFrom, _context_name: str = ''
    ) -> Dict[str, Any]:
        """Convert a ChoiceFrom node (&group or &enum)."""
        # This is typically used for enumerations
        result: Dict[str, Any] = {'type': 'string'}

        if hasattr(choice_from, 'getChildren'):
            for child in choice_from.getChildren():
                if hasattr(child, 'name'):
                    # Reference to an enumeration group
                    normalized_name = avro_name(child.name)
                    return {'$ref': f'#/definitions/{normalized_name}'}

        return result

    def _scan_for_uses(self, structure_schema: Dict[str, Any]) -> List[str]:
        """Scan the structure schema for extension feature usage."""
        uses = set()

        def scan(obj: Any) -> None:
            if isinstance(obj, dict):
                for k, v in obj.items():
                    if k == 'altnames':
                        uses.add('JSONStructureAlternateNames')
                    if k in {'unit', 'currency', 'symbol'}:
                        uses.add('JSONStructureUnits')
                    if k in {'pattern', 'minLength', 'maxLength', 'minimum', 'maximum',
                            'exclusiveMinimum', 'exclusiveMaximum', 'multipleOf',
                            'const', 'enum', 'required', 'minItems', 'maxItems', 'default'}:
                        uses.add('JSONStructureValidation')
                    if k in {'if', 'then', 'else', 'dependentRequired', 'dependentSchemas',
                            'anyOf', 'allOf', 'oneOf', 'not'}:
                        uses.add('JSONStructureConditionalComposition')
                    # Note: $extends is a core keyword per Section 3.10.2, not an add-in
                    scan(v)
            elif isinstance(obj, list):
                for item in obj:
                    scan(item)

        scan(structure_schema)
        return sorted(uses)


def convert_cddl_to_structure(cddl_content: str, namespace: str = DEFAULT_NAMESPACE,
                               validate: bool = True) -> str:
    """
    Convert CDDL content to JSON Structure format.

    Args:
        cddl_content: The CDDL schema as a string
        namespace: The namespace for the schema
        validate: If True, validate the output schema using json-structure SDK

    Returns:
        JSON Structure schema as a string

    Raises:
        CddlConversionError: If validation is enabled and the schema is invalid
    """
    converter = CddlToStructureConverter()
    converter.root_namespace = namespace
    result = converter.convert_cddl_to_structure(cddl_content, namespace)

    # Validate the output schema
    if validate:
        result_str = json.dumps(result, indent=2)
        errors = converter.validate_structure_schema(result, result_str)
        error_errors = [e for e in errors if e.severity == ValidationSeverity.ERROR]
        if error_errors:
            error_messages = '; '.join(f"{e.message} at {e.path}" for e in error_errors)
            raise CddlConversionError(
                f"Generated schema validation failed: {error_messages}",
                context="output validation"
            )

    return json.dumps(result, indent=2)


def convert_cddl_to_structure_files(
    cddl_file_path: str,
    structure_schema_path: str,
    namespace: Optional[str] = None,
    validate: bool = True
) -> None:
    """
    Convert a CDDL file to JSON Structure format.

    Args:
        cddl_file_path: Path to the input CDDL file
        structure_schema_path: Path to the output JSON Structure file
        namespace: Optional namespace for the schema
        validate: If True, validate the output schema using json-structure SDK

    Raises:
        CddlConversionError: If validation is enabled and the schema is invalid
    """
    # Use default namespace if None provided
    if namespace is None:
        namespace = DEFAULT_NAMESPACE

    # Read the CDDL file
    with open(cddl_file_path, 'r', encoding='utf-8') as f:
        cddl_content = f.read()

    # Convert to JSON Structure (validation happens in convert_cddl_to_structure)
    result = convert_cddl_to_structure(cddl_content, namespace, validate)

    # Write the result
    with open(structure_schema_path, 'w', encoding='utf-8') as f:
        f.write(result)



# coding: utf-8
"""
Module to convert Parquet schema to Avro schema.
"""

import json
import os
import pyarrow as pa
import pyarrow.parquet as pq

from avrotize.common import avro_name

class ParquetToAvroConverter:
    """
    Class to convert Parquet schema to Avro schema.
    """

    def __init__(self, parquet_file_path, avro_schema_path, namespace=""):
        """
        Initialize the converter with file paths and namespace.

        :param parquet_file_path: Path to the Parquet file.
        :param avro_schema_path: Path to save the Avro schema file.
        :param namespace: Namespace for Avro records.
        """
        self.parquet_file_path = parquet_file_path
        self.avro_schema_path = avro_schema_path
        self.namespace = namespace

    def convert(self):
        """
        Convert Parquet schema to Avro schema and save to file.
        """
        parquet_table = pq.read_table(self.parquet_file_path)
        schema = parquet_table.schema

        # Infer the name of the schema from the parquet file name
        schema_name = avro_name(os.path.basename(self.parquet_file_path).split(".")[0])

        # Update the avro_schema dictionary
        avro_schema = {
            "type": "record",
            "name": schema_name,
            "namespace": self.namespace,
            "fields": []
        }

        for field in schema:
            avro_field = self.convert_parquet_field_to_avro_field(field)
            avro_schema["fields"].append(avro_field)

        with open(self.avro_schema_path, "w", encoding="utf-8") as file:
            json.dump(avro_schema, file, indent=2)

    def convert_parquet_field_to_avro_field(self, field):
        """
        Convert a Parquet field to an Avro field.

        :param field: Parquet field to convert.
        :return: Avro field as a dictionary.
        """
        avro_type = self.convert_parquet_type_to_avro_type(field.type, field.name)
        avro_field = {
            "name": field.name,
            "type": avro_type
        }
        if field.metadata and b'description' in field.metadata:
            avro_field["doc"] = field.metadata[b'description'].decode("utf-8")
        return avro_field

    def convert_parquet_type_to_avro_type(self, parquet_type, field_name):
        """
        Convert a Parquet type to an Avro type.

        :param parquet_type: Parquet type to convert.
        :param field_name: Name of the field being converted.
        :return: Avro type as a string or dictionary.
        """
        if pa.types.is_int8(parquet_type):
            return "int"
        if pa.types.is_int16(parquet_type):
            return "int"
        if pa.types.is_int32(parquet_type):
            return "int"
        if pa.types.is_int64(parquet_type):
            return "long"
        if pa.types.is_uint8(parquet_type):
            return "int"
        if pa.types.is_uint16(parquet_type):
            return "int"
        if pa.types.is_uint32(parquet_type):
            return "long"
        if pa.types.is_uint64(parquet_type):
            return "long"
        if pa.types.is_float32(parquet_type):
            return "float"
        if pa.types.is_float64(parquet_type):
            return "double"
        if pa.types.is_boolean(parquet_type):
            return "boolean"
        if pa.types.is_binary(parquet_type):
            return "bytes"
        if pa.types.is_string(parquet_type):
            return "string"
        if pa.types.is_timestamp(parquet_type):
            return {"type": "long", "logicalType": "timestamp-millis"}
        if pa.types.is_date32(parquet_type):
            return {"type": "int", "logicalType": "date"}
        if pa.types.is_date64(parquet_type):
            return {"type": "long", "logicalType": "timestamp-millis"}
        if pa.types.is_list(parquet_type):
            return {
                "type": "array",
                "items": self.convert_parquet_type_to_avro_type(parquet_type.value_type, field_name)
            }
        if pa.types.is_map(parquet_type):
            return {
                "type": "map",
                "values": self.convert_parquet_type_to_avro_type(parquet_type.item_type, field_name)
            }
        if pa.types.is_struct(parquet_type):
            fields = [
                {
                    "name": nested_field.name,
                    "type": self.convert_parquet_type_to_avro_type(nested_field.type, nested_field.name)
                } for nested_field in parquet_type
            ]
            return {
                "type": "record",
                "name": f"{field_name}Type",
                "namespace": self.namespace,
                "fields": fields
            }
        if pa.types.is_decimal(parquet_type):
            return {
                "type": "bytes",
                "logicalType": "decimal",
                "precision": parquet_type.precision,
                "scale": parquet_type.scale
            }
        return "string"

def convert_parquet_to_avro(parquet_file_path, avro_file_path, namespace=""):
    """
    Convert a Parquet file to an Avro schema file.

    :param parquet_file_path: Path to the Parquet file.
    :param avro_file_path: Path to save the Avro schema file.
    :param namespace: Namespace for Avro records.
    """
    
    if not os.path.exists(parquet_file_path):
        raise FileNotFoundError(f"Parquet file not found: {parquet_file_path}")
    
    converter = ParquetToAvroConverter(parquet_file_path, avro_file_path, namespace)
    converter.convert()




# coding: utf-8
"""
Module to convert JSON Structure schema to GraphQL schema.
"""

import json
import os
from typing import Dict, List, Optional, Set

from avrotize.common import get_longest_namespace_prefix

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | bool | int | None


class StructureToGraphQLConverter:
    """
    Class to convert JSON Structure schema to GraphQL schema.
    """

    def __init__(self, structure_schema_path, graphql_schema_path):
        """
        Initialize the converter with file paths.

        :param structure_schema_path: Path to the JSON Structure schema file.
        :param graphql_schema_path: Path to save the GraphQL schema file.
        """
        self.structure_schema_path = structure_schema_path
        self.graphql_schema_path = graphql_schema_path
        self.records = {}  # qualified_name -> schema
        self.record_names = {}  # qualified_name -> simple_name
        self.enums = {}  # qualified_name -> schema
        self.enum_names = {}  # qualified_name -> simple_name
        self.scalars = {}
        self.schema_doc: JsonNode = None
        self.definitions: Dict = {}
        self.schema_registry: Dict[str, Dict] = {}
        self.longest_namespace_prefix = ""
        self.type_order = []  # Track order of type definitions for dependency ordering

    def convert(self: 'StructureToGraphQLConverter'):
        """
        Convert JSON Structure schema to GraphQL schema and save to file.
        """
        with open(self.structure_schema_path, 'r', encoding='utf-8') as file:
            structure_schemas: JsonNode = json.load(file)

        # Normalize to list
        if isinstance(structure_schemas, dict):
            structure_schemas = [structure_schemas]
        
        if not isinstance(structure_schemas, list):
            raise ValueError("Expected a single JSON Structure schema as a JSON object, or a list of schemas")

        self.schema_doc = structure_schemas
        
        # Register all schemas with $id
        for schema in structure_schemas:
            if isinstance(schema, dict):
                self.register_schema_ids(schema)

        # Extract named types from all schemas
        for schema in structure_schemas:
            if isinstance(schema, dict):
                # Process root type if it exists
                if 'type' in schema:
                    self.extract_named_types_from_structure(schema, schema.get('namespace', ''))
                elif '$root' in schema:
                    root_ref = schema['$root']
                    root_schema = self.resolve_ref(root_ref, schema)
                    if root_schema:
                        ref_path = root_ref.split('/')
                        ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else ''
                        self.extract_named_types_from_structure(root_schema, ref_namespace)
                
                # Process definitions
                if 'definitions' in schema:
                    self.definitions = schema['definitions']
                    self.process_definitions(self.definitions, '')

        graphql_content = self.generate_graphql()

        with open(self.graphql_schema_path, "w", encoding="utf-8") as file:
            file.write(graphql_content)
            if not graphql_content.endswith('\n'):
                file.write('\n')

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """Recursively registers schemas with $id keywords"""
        if not isinstance(schema, dict):
            return

        if '$id' in schema:
            schema_id = schema['$id']
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id

        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)

        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)

        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """Resolves a $ref to the actual schema definition"""
        if not ref.startswith('#/'):
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None

        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc
        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]
        return schema

    def process_definitions(self, definitions: Dict, namespace_path: str) -> None:
        """Processes the definitions section recursively"""
        for name, definition in definitions.items():
            if isinstance(definition, dict):
                if 'type' in definition or 'enum' in definition:
                    # This is a type definition
                    current_namespace = self.concat_namespace(namespace_path, '')
                    self.extract_named_types_from_structure(definition, current_namespace, explicit_name=name)
                else:
                    # This is a namespace
                    new_namespace = self.concat_namespace(namespace_path, name)
                    self.process_definitions(definition, new_namespace)

    def concat_namespace(self, namespace: str, name: str) -> str:
        """Concatenates namespace and name with a dot separator"""
        if namespace and name:
            return f"{namespace}.{name}"
        elif namespace:
            return namespace
        else:
            return name

    def qualified_name(self, schema: Dict[str, JsonNode], parent_namespace: str = '') -> str:
        """
        Get the full name of a record type.
        """
        name = str(schema.get('name', 'UnnamedType'))
        namespace = str(schema.get('namespace', parent_namespace))
        if namespace:
            return f"{namespace}_{name}".replace('.', '_')
        return name

    def extract_named_types_from_structure(self, schema: Dict, parent_namespace: str, explicit_name: str = ''):
        """
        Extract all named types (objects, enums) from a JSON Structure schema.
        """
        if not isinstance(schema, dict):
            return

    def extract_named_types_from_structure(self, schema: Dict, parent_namespace: str, explicit_name: str = ''):
        """
        Extract all named types (objects, enums) from a JSON Structure schema.
        """
        if not isinstance(schema, dict):
            return

        # Handle $ref FIRST before anything else
        if '$ref' in schema:
            # Use self.schema_doc as context for resolving refs
            context = self.schema_doc[0] if isinstance(self.schema_doc, list) and len(self.schema_doc) > 0 else self.schema_doc
            ref_schema = self.resolve_ref(schema['$ref'], context)
            if ref_schema:
                # Extract type name from $ref path for explicit naming
                ref_path = schema['$ref'].split('/')
                ref_name = ref_path[-1]
                self.extract_named_types_from_structure(ref_schema, parent_namespace, explicit_name=ref_name)
            return

        # Use explicit name if provided, otherwise get from schema
        name = explicit_name if explicit_name else schema.get('name', '')
        
        # Handle enum keyword
        if 'enum' in schema:
            qualified = self.qualified_name({**schema, 'name': name, 'namespace': parent_namespace}, parent_namespace)
            if qualified not in self.enums:
                self.enums[qualified] = schema
                self.enum_names[qualified] = name
                self.type_order.append(('enum', qualified))
            return

        # Handle type keyword
        struct_type = schema.get('type')
        
        if struct_type == 'object':
            # Process nested properties FIRST to ensure dependencies come before this type
            if 'properties' in schema and isinstance(schema['properties'], dict):
                for prop_name, prop_schema in schema['properties'].items():
                    if isinstance(prop_schema, dict):
                        self.extract_named_types_from_structure(prop_schema, parent_namespace)
            
            # NOW add this type after all dependencies have been processed
            if name:
                qualified = self.qualified_name({**schema, 'name': name, 'namespace': parent_namespace}, parent_namespace)
                if qualified not in self.records:
                    self.records[qualified] = schema
                    self.record_names[qualified] = name
                    self.type_order.append(('record', qualified))
        
        elif struct_type == 'array' and 'items' in schema:
            if isinstance(schema['items'], dict):
                self.extract_named_types_from_structure(schema['items'], parent_namespace)
        
        elif struct_type == 'set' and 'items' in schema:
            if isinstance(schema['items'], dict):
                self.extract_named_types_from_structure(schema['items'], parent_namespace)
        
        elif struct_type == 'map' and 'values' in schema:
            if isinstance(schema['values'], dict):
                self.extract_named_types_from_structure(schema['values'], parent_namespace)
        
        elif struct_type == 'choice':
            # Process choice types
            choices = schema.get('choices', {})
            for choice_name, choice_schema in choices.items():
                if isinstance(choice_schema, dict):
                    if '$ref' in choice_schema:
                        ref_schema = self.resolve_ref(choice_schema['$ref'], schema)
                        if ref_schema:
                            self.extract_named_types_from_structure(ref_schema, parent_namespace)
                    else:
                        self.extract_named_types_from_structure(choice_schema, parent_namespace)

    def generate_graphql(self):
        """
        Generate GraphQL content from the extracted types.

        :return: GraphQL content as a string.
        """
        graphql = []

        # Generate scalars for custom types
        custom_scalars = set()
        
        # Add commonly used custom scalars
        if any('date' in str(record).lower() or 'datetime' in str(record).lower() or 'timestamp' in str(record).lower() 
               for record in self.records.values()):
            custom_scalars.add('scalar Date')
            custom_scalars.add('scalar DateTime')
        
        if any('uuid' in str(record).lower() for record in self.records.values()):
            custom_scalars.add('scalar UUID')
        
        if any('uri' in str(record).lower() or 'url' in str(record).lower() for record in self.records.values()):
            custom_scalars.add('scalar URI')
        
        if any('decimal' in str(record).lower() for record in self.records.values()):
            custom_scalars.add('scalar Decimal')
        
        if any('binary' in str(record).lower() or 'bytes' in str(record).lower() for record in self.records.values()):
            custom_scalars.add('scalar Binary')
        
        # Add JSON scalar for map types
        custom_scalars.add('scalar JSON')
        
        for scalar in sorted(custom_scalars):
            graphql.append(scalar)
        
        if custom_scalars:
            graphql.append('')  # Empty line after scalars

        # Generate types in dependency order
        for type_kind, qualified_name in self.type_order:
            if type_kind == 'enum' and qualified_name in self.enums:
                graphql.append(self.generate_graphql_enum(self.enums[qualified_name]))
            elif type_kind == 'record' and qualified_name in self.records:
                graphql.append(self.generate_graphql_record(self.records[qualified_name]))

        return "\n".join(graphql)

    def generate_graphql_record(self, record):
        """
        Generate GraphQL content for a record.

        :param record: Record schema as a dictionary.
        :return: GraphQL content as a string.
        """
        name = record.get('name', 'UnnamedType')
        doc = record.get('description', record.get('doc', ''))
        
        # Add description as comment if present
        output = []
        if doc:
            output.append(f'"""\n{doc}\n"""')
        
        fields = []
        properties = record.get('properties', {})
        required_props = record.get('required', [])
        
        for prop_name, prop_schema in properties.items():
            field_type = self.get_graphql_type(prop_schema)
            
            # Add ! for required fields
            is_required = prop_name in required_props if isinstance(required_props, list) else False
            if is_required and not field_type.endswith('!'):
                field_type = f"{field_type}!"
            
            # Add field description as comment if present
            field_doc = prop_schema.get('description', prop_schema.get('doc', ''))
            if field_doc:
                fields.append(f'  """{field_doc}"""')
            fields.append(f"  {prop_name}: {field_type}")
        
        output.append(f"type {name} {{")
        if fields:
            output.extend(fields)
        else:
            output.append("  _empty: String")  # GraphQL doesn't allow empty types
        output.append("}")
        
        return "\n".join(output)

    def generate_graphql_enum(self, enum):
        """
        Generate GraphQL content for an enum.

        :param enum: Enum schema as a dictionary.
        :return: GraphQL content as a string.
        """
        name = enum.get('name', 'UnnamedEnum')
        doc = enum.get('description', enum.get('doc', ''))
        symbols = enum.get('enum', [])
        
        output = []
        if doc:
            output.append(f'"""\n{doc}\n"""')
        
        # GraphQL enum members must be valid identifiers
        enum_members = []
        for symbol in symbols:
            # Convert to valid GraphQL identifier
            member = str(symbol).replace('-', '_').replace(' ', '_').replace('.', '_')
            # Ensure it starts with a letter or underscore
            if member and not member[0].isalpha() and member[0] != '_':
                member = f"_{member}"
            enum_members.append(f"  {member}")
        
        output.append(f"enum {name} {{")
        if enum_members:
            output.extend(enum_members)
        else:
            output.append("  _EMPTY")
        output.append("}")
        
        return "\n".join(output)

    def get_graphql_type(self, structure_type):
        """
        Get GraphQL type from JSON Structure type.

        :param structure_type: JSON Structure type as a string, dict, or list.
        :return: GraphQL type as a string.
        """
        if isinstance(structure_type, list):
            # Handle type unions (e.g., ["null", "string"])
            non_null_types = [t for t in structure_type if t != "null"]
            if non_null_types:
                return self.get_graphql_type(non_null_types[0])
            return "String"
        
        if isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                # Use self.schema_doc as context for resolving refs
                context = self.schema_doc[0] if isinstance(self.schema_doc, list) and len(self.schema_doc) > 0 else self.schema_doc
                ref_schema = self.resolve_ref(structure_type['$ref'], context)
                if ref_schema:
                    # First try to get the name from the schema
                    ref_name = ref_schema.get('name')
                    if ref_name:
                        return ref_name
                    # Try to extract name from $ref path
                    ref_path = structure_type['$ref'].split('/')
                    # The last element is the type name
                    type_name = ref_path[-1]
                    return type_name
                return "JSON"
            
            # Handle enum
            if 'enum' in structure_type:
                # Return inline enum as String (GraphQL doesn't support inline enums)
                return "String"
            
            # Handle type keyword
            struct_type = structure_type.get('type')
            
            # Handle type unions (e.g., ["string", "null"])
            if isinstance(struct_type, list):
                return self.get_graphql_type(struct_type)
            
            if struct_type == 'array':
                items_type = self.get_graphql_type(structure_type.get('items', {'type': 'any'}))
                return f"[{items_type}]"
            
            if struct_type == 'set':
                items_type = self.get_graphql_type(structure_type.get('items', {'type': 'any'}))
                return f"[{items_type}]"  # GraphQL doesn't have Set, use array
            
            if struct_type == 'map':
                return "JSON"  # GraphQL doesn't have Map, use custom scalar
            
            if struct_type == 'object':
                # Inline object - return JSON scalar
                return "JSON"
            
            if struct_type == 'choice':
                # Union types in GraphQL are more complex, return JSON for now
                return "JSON"
            
            if struct_type == 'tuple':
                # Tuples are represented as arrays in GraphQL
                return "JSON"
            
            if struct_type:
                return self.get_graphql_primitive_type(struct_type)
        
        if isinstance(structure_type, str):
            return self.get_graphql_primitive_type(structure_type)
        
        return "JSON"

    def get_graphql_primitive_type(self, structure_type):
        """
        Map JSON Structure primitive types to GraphQL types.

        :param structure_type: JSON Structure type as a string.
        :return: GraphQL type as a string.
        """
        type_mapping = {
            # JSON primitive types
            "string": "String",
            "number": "Float",
            "integer": "Int",
            "boolean": "Boolean",
            "null": "String",
            
            # Extended integer types
            "int8": "Int",
            "uint8": "Int",
            "int16": "Int",
            "uint16": "Int",
            "int32": "Int",
            "uint32": "Int",
            "int64": "Int",
            "uint64": "Int",
            "int128": "String",  # GraphQL Int is 32-bit, use String for larger
            "uint128": "String",
            
            # Floating point types
            "float8": "Float",
            "float": "Float",
            "double": "Float",
            "binary32": "Float",
            "binary64": "Float",
            "decimal": "Decimal",
            
            # Binary types
            "binary": "Binary",
            
            # Date/time types
            "date": "Date",
            "time": "String",
            "datetime": "DateTime",
            "timestamp": "DateTime",
            "duration": "String",
            
            # Other special types
            "uuid": "UUID",
            "uri": "URI",
            "jsonpointer": "String",
            "any": "JSON"
        }
        
        return type_mapping.get(structure_type, 'String')


def convert_structure_to_graphql(structure_schema_path, graphql_schema_path):
    """
    Convert a JSON Structure schema file to a GraphQL schema file.

    :param structure_schema_path: Path to the JSON Structure schema file.
    :param graphql_schema_path: Path to save the GraphQL schema file.
    """
    converter = StructureToGraphQLConverter(structure_schema_path, graphql_schema_path)
    converter.convert()



"""Converts an Avro schema to a Kusto table schema."""

import json
import sys
from typing import Any, List
from avrotize.common import build_flat_type_dict, inline_avro_references, strip_first_doc
from azure.kusto.data import KustoClient, KustoConnectionStringBuilder, ClientRequestProperties


class AvroToKusto:
    """Converts an Avro schema to a Kusto table schema."""

    def ___init___(self):
        """Initializes a new instance of the AvroToKusto class."""
        pass

    def convert_record_to_kusto(self, type_dict:dict, recordschema: dict, emit_cloudevents_columns: bool, emit_cloudevents_dispatch_table: bool) -> List[str]:
        """Converts an Avro record schema to a Kusto table schema."""
        # Get the name and fields of the top-level record
        table_name = recordschema["name"]
        fields = recordschema["fields"]
        
        # Create a StringBuilder to store the kusto statements
        kusto = []

        # Append the create table statement with the column names and types
        kusto.append(f".create-merge table [{table_name}] (")
        columns = []
        for field in fields:
            column_name = field["name"]
            column_type = self.convert_avro_type_to_kusto_type(field["type"])
            columns.append(f"   [{column_name}]: {column_type}")
        if emit_cloudevents_columns:
            columns.append("   [___type]: string")
            columns.append("   [___source]: string")
            columns.append("   [___id]: string")
            columns.append("   [___time]: datetime")
            columns.append("   [___subject]: string")
        kusto.append(",\n".join(columns))
        kusto.append(");")
        kusto.append("")

        # Add the doc string as table metadata
        if "doc" in recordschema:
            doc_data = recordschema["doc"]
            doc_data = (doc_data[:997] + "...") if len(doc_data) > 1000 else doc_data
            doc_string = json.dumps(json.dumps({
                "description": doc_data
            }))
            kusto.append(
                f".alter table [{table_name}] docstring {doc_string};")
            kusto.append("")

        doc_string_statement = []
        for field in fields:
            column_name = field["name"]
            if "doc" in field:
                doc_data = field["doc"]
                if len(doc_data) > 900:
                    doc_data = (doc_data[:897] + "...")
                doc_content = {
                    "description": doc_data
                }
                if isinstance(field["type"], list) or isinstance(field["type"], dict):
                    inline_schema = inline_avro_references(field["type"].copy(), type_dict.copy(), '')
                    if (len(json.dumps(inline_schema)) + len(doc_data)) > 900:
                        while strip_first_doc(inline_schema):
                            if (len(json.dumps(inline_schema)) + len(doc_data)) < 900:
                                break
                        if (len(json.dumps(inline_schema)) + len(doc_data)) > 900:
                            doc_content["schema"] = '{ "doc": "Schema too large to inline. Please refer to the Avro schema for more details." }'
                        else:
                            doc_content["schema"] = inline_schema
                    else:
                        doc_content["schema"] = inline_schema
                doc = json.dumps(json.dumps(doc_content))
                doc_string_statement.append(f"   [{column_name}]: {doc}")
        if doc_string_statement and emit_cloudevents_columns:
            doc_string_statement.extend([
                "   [___type] : 'Event type'",
                "   [___source]: 'Context origin/source of the event'",
                "   [___id]: 'Event identifier'",
                "   [___time]: 'Event generation time'",
                "   [___subject]: 'Context subject of the event'"
            ])
        if doc_string_statement:
            kusto.append(f".alter table [{table_name}] column-docstrings (")
            kusto.append(",\n".join(doc_string_statement))
            kusto.append(");")
            kusto.append("")

        # add the JSON mapping for the table
        # .create-or-alter table dfl_data_events ingestion json mapping
        kusto.append(
            f".create-or-alter table [{table_name}] ingestion json mapping \"{table_name}_json_flat\"")
        kusto.append("```\n[")
        if emit_cloudevents_columns:
            kusto.append("  {\"column\": \"___type\", \"path\": \"$.type\"},")
            kusto.append(
                "  {\"column\": \"___source\", \"path\": \"$.source\"},")
            kusto.append("  {\"column\": \"___id\", \"path\": \"$.id\"},")
            kusto.append("  {\"column\": \"___time\", \"path\": \"$.time\"},")
            kusto.append(
                "  {\"column\": \"___subject\", \"path\": \"$.subject\"},")
        for field in fields:
            json_name = column_name = field["name"]
            if 'altnames' in field:
                if 'kql' in field['altnames']:
                    column_name = field['altnames']['kql']
                if 'json' in field['altnames']:
                    json_name = field['altnames']['json']
            kusto.append(
                f"  {{\"column\": \"{column_name}\", \"path\": \"$.{json_name}\"}},")
        kusto.append("]\n```\n\n")

        if emit_cloudevents_columns:
            kusto.append(
                f".create-or-alter table [{table_name}] ingestion json mapping \"{table_name}_json_ce_structured\"")
            kusto.append("```\n[")
            kusto.append("  {\"column\": \"___type\", \"path\": \"$.type\"},")
            kusto.append(
                "  {\"column\": \"___source\", \"path\": \"$.source\"},")
            kusto.append("  {\"column\": \"___id\", \"path\": \"$.id\"},")
            kusto.append("  {\"column\": \"___time\", \"path\": \"$.time\"},")
            kusto.append(
                "  {\"column\": \"___subject\", \"path\": \"$.subject\"},")
            for field in fields:
                json_name = column_name = field["name"]
                if 'altnames' in field:
                    if 'kql' in field['altnames']:
                        column_name = field['altnames']['kql']
                    if 'json' in field['altnames']:
                        json_name = field['altnames']['json']
                kusto.append(
                    f"  {{\"column\": \"{column_name}\", \"path\": \"$.data.{json_name}\"}},")
            kusto.append("]\n```\n\n")

        if emit_cloudevents_columns:
            kusto.append(
                f".drop materialized-view {table_name}Latest ifexists;")
            kusto.append("")
            kusto.append(
                f".create materialized-view with (backfill=true) {table_name}Latest on table {table_name} {{")
            kusto.append(
                f"    {table_name} | summarize arg_max(___time, *) by ___type, ___source, ___subject")
            kusto.append("}")
            kusto.append("")

        if emit_cloudevents_dispatch_table:
            event_type = recordschema["namespace"] + "." + \
                recordschema["name"] if "namespace" in recordschema else recordschema["name"]

            query = f"_cloudevents_dispatch | where (specversion == '1.0' and type == '{event_type}') | " + \
                    "project"                        
            for field in fields:
                column_name = field["name"]
                if "altnames" in field and "kql" in field["altnames"]:
                    column_name = field["altnames"]["kql"]
                column_type = self.convert_avro_type_to_kusto_type(
                    field["type"])
                query += f"['{column_name}'] = to{column_type}(data.['{column_name}']),"
            query += "___type = type,___source = source,___id = ['id'],___time = ['time'],___subject = subject"

            # build an update policy for the table that gets triggered by updates to the dispatch table and extracts the event
            kusto.append(f".alter table [{table_name}] policy update")
            kusto.append("```")
            kusto.append("[{")
            kusto.append("  \"IsEnabled\": true,")
            kusto.append("  \"Source\": \"_cloudevents_dispatch\",")
            kusto.append(
                f"  \"Query\": \"{query}\",")
            kusto.append("  \"IsTransactional\": false,")
            kusto.append("  \"PropagateIngestionProperties\": true,")
            kusto.append("}]")
            kusto.append("```\n")

        return kusto

    def convert_avro_to_kusto_script(self, avro_schema_path, avro_record_type, emit_cloudevents_columns=False, emit_cloudevents_dispatch_table=False) -> str:
        """Converts an Avro schema to a Kusto table schema."""
        if emit_cloudevents_dispatch_table:
            emit_cloudevents_columns = True
        schema_file = avro_schema_path
        if not schema_file:
            print("Please specify the avro schema file")
            sys.exit(1)
        with open(schema_file, "r", encoding="utf-8") as f:
            schema_json = f.read()

        # Parse the schema as a JSON object
        schema = json.loads(schema_json)

        if isinstance(schema, list):
            if avro_record_type:
                schema = next(
                    (x for x in schema if x["name"] == avro_record_type), None)
                if schema is None:
                    print(
                        f"No record type {avro_record_type} found in the Avro schema")
                    sys.exit(1)
        elif not isinstance(schema, dict) or "type" not in schema or schema["type"] != "record":
            print(
                "Expected a single Avro schema as a JSON object, or a list of schema records")
            sys.exit(1)

        if not isinstance(schema, list):
            schema = [schema]

        kusto_script = []

        if emit_cloudevents_dispatch_table:
            kusto_script.append(
                ".create-merge table [_cloudevents_dispatch] (")
            kusto_script.append("    [specversion]: string,")
            kusto_script.append("    [type]: string,")
            kusto_script.append("    [source]: string,")
            kusto_script.append("    [id]: string,")
            kusto_script.append("    [time]: datetime,")
            kusto_script.append("    [subject]: string,")
            kusto_script.append("    [datacontenttype]: string,")
            kusto_script.append("    [dataschema]: string,")
            kusto_script.append("    [data]: dynamic")
            kusto_script.append(");\n\n")
            kusto_script.append(
                ".create-or-alter table [_cloudevents_dispatch] ingestion json mapping \"_cloudevents_dispatch_json\"")
            kusto_script.append("```\n[")
            kusto_script.append(
                "  {\"column\": \"specversion\", \"path\": \"$.specversion\"},")
            kusto_script.append(
                "  {\"column\": \"type\", \"path\": \"$.type\"},")
            kusto_script.append(
                "  {\"column\": \"source\", \"path\": \"$.source\"},")
            kusto_script.append("  {\"column\": \"id\", \"path\": \"$.id\"},")
            kusto_script.append(
                "  {\"column\": \"time\", \"path\": \"$.time\"},")
            kusto_script.append(
                "  {\"column\": \"subject\", \"path\": \"$.subject\"},")
            kusto_script.append(
                "  {\"column\": \"datacontenttype\", \"path\": \"$.datacontenttype\"},")
            kusto_script.append(
                "  {\"column\": \"dataschema\", \"path\": \"$.dataschema\"},")
            kusto_script.append(
                "  {\"column\": \"data\", \"path\": \"$.data\"}")
            kusto_script.append("]\n```\n\n")

        type_dict =  build_flat_type_dict(schema)
        for record in schema:
            if not isinstance(record, dict) or "type" not in record or record["type"] != "record":
                continue
            kusto_script.extend(self.convert_record_to_kusto(type_dict,
                record, emit_cloudevents_columns, emit_cloudevents_dispatch_table))
        return "\n".join(kusto_script)

    def convert_avro_to_kusto_file(self, avro_schema_path, avro_record_type, kusto_file_path, emit_cloudevents_columns=False, emit_cloudevents_dispatch_table=False):
        """Converts an Avro schema to a Kusto table schema."""
        script = self.convert_avro_to_kusto_script(
            avro_schema_path, avro_record_type, emit_cloudevents_columns, emit_cloudevents_dispatch_table)
        with open(kusto_file_path, "w", encoding="utf-8") as kusto_file:
            kusto_file.write(script)

    def convert_avro_type_to_kusto_type(self, avro_type: str | dict | list):
        """Converts an Avro type to a Kusto type."""
        if isinstance(avro_type, list):
            # If the type is an array, then it is a union type. Look whether it's a pair of a scalar type and null:
            itemCount = len(avro_type)
            if itemCount > 2:
                return "dynamic"
            if itemCount == 1:
                return self.convert_avro_type_to_kusto_type(avro_type[0])
            else:
                first = avro_type[0]
                second = avro_type[1]
                if isinstance(first, str) and first == "null":
                    return self.convert_avro_type_to_kusto_type(second)
                elif isinstance(second, str) and second == "null":
                    return self.convert_avro_type_to_kusto_type(first)
                else:
                    return "dynamic"
        elif isinstance(avro_type, dict):
            type_value = avro_type.get("type")
            if type_value == "enum":
                return "string"
            elif type_value == "fixed":
                return "dynamic"
            elif type_value == "string":
                logical_type = avro_type.get("logicalType")
                if logical_type == "uuid":
                    return "guid"
                return "string"
            elif type_value == "bytes":
                logical_type = avro_type.get("logicalType")
                if logical_type == "decimal":
                    return "decimal"
                return "dynamic"
            elif type_value == "long":
                logical_type = avro_type.get("logicalType")
                if logical_type in ["timestamp-millis", "timestamp-micros"]:
                    return "datetime"
                if logical_type in ["time-millis", "time-micros"]:
                    return "timespan"
                return "long"
            elif type_value == "int":
                logical_type = avro_type.get("logicalType")
                if logical_type == "date":
                    return "datetime"
                return "int"
            else:
                return self.map_scalar_type(type_value)
        elif isinstance(avro_type, str):
            return self.map_scalar_type(avro_type)

    def map_scalar_type(self, type_value: Any):
        """Maps an Avro scalar type to a Kusto scalar type."""
        if type_value == "null":
            return "dynamic"
        elif type_value == "int":
            return "int"
        elif type_value == "long":
            return "long"
        elif type_value == "float":
            return "real"
        elif type_value == "double":
            return "real"
        elif type_value == "boolean":
            return "bool"
        elif type_value == "bytes":
            return "dynamic"
        elif type_value == "string":
            return "string"
        else:
            return "dynamic"


def convert_avro_to_kusto_file(avro_schema_path, avro_record_type, kusto_file_path, emit_cloudevents_columns=False, emit_cloudevents_dispatch_table=False):
    """Converts an Avro schema to a Kusto table schema."""
    avro_to_kusto = AvroToKusto()
    avro_to_kusto.convert_avro_to_kusto_file(
        avro_schema_path, avro_record_type, kusto_file_path, emit_cloudevents_columns, emit_cloudevents_dispatch_table)


def convert_avro_to_kusto_db(avro_schema_path, avro_record_type, kusto_uri, kusto_database, emit_cloudevents_columns=False, emit_cloudevents_dispatch_table=False, token_provider=None):
    """Converts an Avro schema to a Kusto table schema."""
    avro_to_kusto = AvroToKusto()
    script = avro_to_kusto.convert_avro_to_kusto_script(
        avro_schema_path, avro_record_type, emit_cloudevents_columns, emit_cloudevents_dispatch_table)
    kcsb = KustoConnectionStringBuilder.with_az_cli_authentication(
        kusto_uri) if not token_provider else KustoConnectionStringBuilder.with_token_provider(kusto_uri, token_provider)
    client = KustoClient(kcsb)
    for statement in script.split("\n\n"):
        if statement.strip():
            try:
                client.execute_mgmt(kusto_database, statement)
            except Exception as e:
                print(e)
                sys.exit(1)

def convert_avro_to_kusto(avro_schema_path, avro_record_type, kusto_file_path, kusto_uri, kusto_database, emit_cloudevents_columns=False, emit_cloudevents_dispatch_table=False, token_provider=None):
    """Converts an Avro schema to a Kusto table schema."""
    if not kusto_uri and not kusto_database:
        convert_avro_to_kusto_file(
            avro_schema_path, avro_record_type, kusto_file_path, emit_cloudevents_columns, emit_cloudevents_dispatch_table)
    else:
        convert_avro_to_kusto_db(
            avro_schema_path, avro_record_type, kusto_uri, kusto_database, emit_cloudevents_columns, emit_cloudevents_dispatch_table, token_provider)